from __future__ import annotations

import datetime
import os
import unittest

from azul_bedrock import models_network as azm

from azul_runner import (
    EventData,
    Feature,
    FeatureValue,
    Job,
    JobResult,
    State,
    TestPlugin,
    add_settings,
    local,
    monitor,
    network,
    network_transform,
)

from . import plugin_support as sup


class TestLoadFile(TestPlugin):
    PLUGIN_TO_TEST = sup.DummyPlugin

    def test_load_cart(self):
        """Ensure the unwrap helper function works as expected on a CaRT file."""
        data = self.load_cart("u_testfile.cart", description="test_file")
        self.assertEqual(data, b"file contents")

    def test_load_test_file_bytes(self):
        """Ensure the unwrap helper function works as expected on a CaRT file."""
        data = self.load_test_file_bytes(
            "7bb6f9f7a47a63e684925af3608c059edcc371eb81188c48c9714896fb1091fd",
            "Small carted test file to verify a file can be loaded.",
        )
        self.assertEqual(data, b"file contents")

    def test_load_file_path(self):
        """Ensure the unwrap helper function works as expected on a CaRT file."""
        path = self.load_test_file_path(
            "7bb6f9f7a47a63e684925af3608c059edcc371eb81188c48c9714896fb1091fd",
            "Small carted test file to verify a file can be loaded.",
        )
        with path.open("rb") as f:
            self.assertEqual(f.read(), b"file contents")


class TestAssertReprEqual(TestPlugin):
    PLUGIN_TO_TEST = sup.DummyPlugin

    def test_formatted(self):
        try:
            self.assertReprEqual(JobResult(state=State(State.Label.OPT_OUT)), JobResult(state=State()))
        except AssertionError as e:
            print(e.args[0])
            self.assertIn(
                r"""'JobResult(state=State(State.Label.OPT_OUT, message="No opt-out reason was provided."))\n' != 'JobResult(state=State(State.Label.COMPLETED))\n'
- JobResult(state=State(State.Label.OPT_OUT, message="No opt-out reason was provided."))
+ JobResult(state=State(State.Label.COMPLETED))
""",
                e.args[0],
            )


class QuickFailPlugin(sup.DummyPlugin):
    """Class definied at module level because add_settings creates a dynamic class which is harder for dill to work."""

    SETTINGS = add_settings(heartbeat_interval=1)


class TestBasePluginStatic(unittest.TestCase):
    """
    Test cases for base plugin class - cases that don't require mock server
    """

    class DPFeatureInheritanceInstance(sup.DummyPluginFeatureInheritance):
        def execute(self, job):
            self.add_feature_values("example_string", ["String value"])
            self.add_feature_values("descendant feature", ["Whatever"])

    def test_feature_inheritance_instance(self):
        """Tests that feature inheritance functions correctly on instantiated classes"""
        loop = monitor.Monitor(self.DPFeatureInheritanceInstance, {})
        self.assertEqual(loop._plugin.VERSION, "2.0")
        self.assertEqual(loop._plugin.cfg.filter_data_types, {})
        self.assertEqual(
            loop._plugin.FEATURES,
            sorted(
                set(sup.DummyPlugin().FEATURES).union(
                    [Feature("descendant feature", "A feature added by the child plugin")]
                )
            ),
        )

        # Check that features give correct output on execution
        entity = azm.BinaryEvent.Entity(sha256="id", datastreams=[], features=[], info={})
        res = loop.run_once(local.gen_event(entity))[None]
        self.assertEqual(res.state, State(State.Label.COMPLETED))
        # Ensure that it returns the features correctly (which means it has passed validation/processing post-run)
        # One feature each from parent and child.
        self.assertEqual(
            res.events[0].features,
            {
                "example_string": [FeatureValue("String value")],
                "descendant feature": [FeatureValue("Whatever")],
            },
        )

    class DPAddNoFeatures(sup.DummyPlugin):
        def execute(self, job):
            pass

    def test_add_no_features(self):
        """Tests that feature inheritance functions correctly on instantiated classes"""
        loop = monitor.Monitor(self.DPAddNoFeatures, {})
        self.assertEqual(loop._plugin.cfg.filter_data_types, {})

        # Check that features give correct output on execution
        entity = azm.BinaryEvent.Entity(sha256="id", datastreams=[], features=[], info={})
        res = loop.run_once(local.gen_event(entity))[None]
        self.assertEqual(res.state, State(State.Label.COMPLETED_EMPTY))

    def test_config_checks(self) -> None:
        self.maxDiff = None
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            sup.DummyPluginMinimum(config=42)
        p = sup.DummyPluginMinimum()
        print(p.cfg.model_dump())
        self.assertEqual(
            p.cfg.model_dump(),
            {
                "heartbeat_interval": 30,
                "run_timeout": 600,
                "max_timeouts_before_exit": 100,
                "server": "",
                "events_url": "",
                "data_url": "",
                "request_timeout": 15,
                "require_expedite": True,
                "require_live": True,
                "require_historic": True,
                "request_retry_count": 3,
                "content_meta_cache_limit": 0,
                "max_values_per_feature": 1000,
                "max_value_length": 4000,
                "plugin_depth_limit": 10,
                "not_ready_backoff": 5,
                "name_remove_prefix": "AzulPlugin",
                "name_suffix": "",
                "version_suffix": "",
                "security_override": "",
                "assume_streams_available": False,
                "deployment_key": "",
                "watch_path": "",
                "watch_type": "",
                "watch_wait": 10,
                "enable_mem_limits": False,
                "used_mem_warning_frac": 0.8,
                "used_mem_force_exit_frac": 0.9,
                "max_mem_file_path": "/sys/fs/cgroup/memory.max",
                "cur_mem_file_path": "/sys/fs/cgroup/memory.current",
                "cur_mem_summary_file_path": "/sys/fs/cgroup/memory.stat",
                "mem_poll_frequency_milliseconds": 1000,
                "filter_max_content_size": 209715200,
                "filter_min_content_size": 0,
                "filter_allow_event_types": [],
                "filter_self": False,
                "filter_data_types": {},
                "concurrent_plugin_instances": 1,
            },
        )

    def test_config_env(self) -> None:
        os.environ["plugin_heartbeat_interval"] = "50"
        # custom plugin option
        os.environ["plugin_myvalue2"] = "100"
        os.environ["plugin_myvalue3"] = "100"

        class CustomPlugin(sup.DummyPluginMinimum):
            SETTINGS = add_settings(myvalue2=(str, "100"), myvalue3=(int, 10))

        p = CustomPlugin(config={"name_suffix": "test", "custom1": 999})
        os.environ.pop("plugin_heartbeat_interval")
        os.environ.pop("plugin_myvalue2")
        print("Actual", p.cfg.model_dump())
        self.assertEqual(
            p.cfg.model_dump(),
            {
                "heartbeat_interval": 50,
                "run_timeout": 600,
                "max_timeouts_before_exit": 100,
                "server": "",
                "events_url": "",
                "data_url": "",
                "request_timeout": 15,
                "require_expedite": True,
                "require_live": True,
                "require_historic": True,
                "request_retry_count": 3,
                "content_meta_cache_limit": 0,
                "max_values_per_feature": 1000,
                "max_value_length": 4000,
                "plugin_depth_limit": 10,
                "not_ready_backoff": 5,
                "name_remove_prefix": "AzulPlugin",
                "name_suffix": "test",
                "version_suffix": "",
                "security_override": "",
                "assume_streams_available": False,
                "deployment_key": "",
                "watch_path": "",
                "watch_type": "",
                "watch_wait": 10,
                "enable_mem_limits": False,
                "used_mem_warning_frac": 0.8,
                "used_mem_force_exit_frac": 0.9,
                "max_mem_file_path": "/sys/fs/cgroup/memory.max",
                "cur_mem_file_path": "/sys/fs/cgroup/memory.current",
                "cur_mem_summary_file_path": "/sys/fs/cgroup/memory.stat",
                "mem_poll_frequency_milliseconds": 1000,
                "filter_max_content_size": 209715200,
                "filter_min_content_size": 0,
                "filter_allow_event_types": [],
                "filter_self": False,
                "filter_data_types": {},
                "concurrent_plugin_instances": 1,
                "custom1": 999,
                "myvalue2": "100",
                "myvalue3": 100,
            },
        )

    def test_config_inheritance(self) -> None:
        # Test that parent/template plugins' default configs are picked up, but overridden by children
        ppt = type(
            "ParentPlugin",
            (sup.DummyPluginMinimum,),
            {"SETTINGS": add_settings(conf1=(int, 1), conf2=(str, "parent"))},
        )
        cpt = type("ChildPlugin", (ppt,), {"SETTINGS": add_settings(conf2=(str, "child"))})
        inst = cpt()
        self.assertEqual(inst.cfg.conf1, 1)
        self.assertEqual(inst.cfg.conf2, "child")
        # Test that explicitly given values override defaults
        inst = cpt(config={"conf2": "override"})
        self.assertEqual(inst.cfg.conf2, "override")

    def test_server_none_errors(self) -> None:
        loop = monitor.Monitor(QuickFailPlugin, {"server": ""})
        net = network.Network(loop._plugin)
        self.assertRaises(ValueError, net.fetch_job)
        self.assertRaises(monitor.TerminateError, loop.run_loop)

    def test_base_class_no_execute(self) -> None:
        p = sup.DummyPluginNoExecute()
        with self.assertRaises(NotImplementedError):
            event = azm.BinaryEvent(
                model_version=azm.CURRENT_MODEL_VERSION,
                kafka_key="runner-placeholder",
                action=azm.BinaryAction.Enriched,
                timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                source=azm.Source(
                    name="source",
                    path=[
                        azm.PathNode(
                            author=azm.Author(name="foo"),
                            action=azm.BinaryAction.Enriched,
                            timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                            sha256="test_entity",
                        ),
                    ],
                    timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                ),
                author=azm.Author(name="TestServer", category="blah"),
                entity=azm.BinaryEvent.Entity(sha256="test_entity", datastreams=[], features=[], info={}),
            )
            p.execute(Job(event=event))

    def test_content_to_API_format(self):
        p = sup.DummyPluginMinimum()
        net = network.Network(p)
        net._cached_file_data = {
            "dummy_hash": azm.Datastream(
                identify_version=1,
                label=azm.DataLabel.CONTENT,
                size=0,
                sha256="256",
                sha1="1",
                md5="5",
                sha512="512",
                mime="mt",
                magic="mm",
                file_format="ft",
            )
        }
        p.data = {"dummy_hash": b"data"}
        self.assertEqual(
            network_transform._to_api_content(
                net._cached_file_data, [EventData(**{"label": azm.DataLabel.TEST, "hash": "dummy_hash"})]
            ),
            [
                azm.Datastream(
                    identify_version=1,
                    label=azm.DataLabel.TEST,
                    size=0,
                    sha256="256",
                    sha1="1",
                    md5="5",
                    sha512="512",
                    mime="mt",
                    magic="mm",
                    file_format="ft",
                )
            ],
        )

    def test_features_to_API_format(self):
        class DP(sup.DummyPlugin):
            FEATURES = [
                Feature(name="another feature", desc="", type=azm.FeatureType.String),
                Feature(name="feature1", desc="", type=azm.FeatureType.String),
            ]

        p = DP()
        self.assertEqual(
            network_transform._to_api_features(
                p,
                {
                    "feature1": {
                        FeatureValue("foo", label="label1"),
                        FeatureValue("bar", label="label2", offset=99),
                        FeatureValue("bar", label="label"),
                    },
                    "another feature": {
                        FeatureValue("whatever"),
                        FeatureValue("404"),
                    },
                },
            ),
            [
                azm.FeatureValue(name="another feature", type="string", value="404"),
                azm.FeatureValue(name="another feature", type="string", value="whatever"),
                azm.FeatureValue(name="feature1", type="string", value="bar", label="label"),
                azm.FeatureValue(name="feature1", type="string", value="bar", label="label2", offset=99),
                azm.FeatureValue(name="feature1", type="string", value="foo", label="label1"),
            ],
        )

    def test_generate_filter_string(self):
        class P(sup.DummyPlugin):
            SETTINGS = add_settings(
                filter_data_types={"example": ["content-type-1", "content-type-2"]},
                filter_max_content_size=500,
                filter_min_content_size=1,
            )

        p = P()
        self.assertEqual(p.cfg.filter_max_content_size, 500)
        self.assertEqual(p.cfg.filter_min_content_size, 1)
        self.assertEqual(p.cfg.filter_data_types, {"example": ["content-type-1", "content-type-2"]})

    def test_generate_wildcard_filter_string(self):
        class P(sup.DummyPlugin):
            SETTINGS = add_settings(
                filter_data_types={"*": ["content-type-1", "content-type-2"]},
                filter_max_content_size=500,
                filter_min_content_size=1,
            )

        p = P()
        self.assertEqual(p.cfg.filter_max_content_size, 500)
        self.assertEqual(p.cfg.filter_min_content_size, 1)
        self.assertEqual(p.cfg.filter_data_types, {"*": ["content-type-1", "content-type-2"]})



import datetime
import unittest

from azul_bedrock import models_network as azm
from azul_bedrock import exceptions_bedrock


class TestFeatureValue(unittest.TestCase):
    def test_value_decode(self):
        v = azm.FeatureValue(name="", type=azm.FeatureType.Integer, value="100")
        self.assertEqual(v.decode_value(), 100)
        v = azm.FeatureValue(name="", type=azm.FeatureType.Float, value="10.0")
        self.assertEqual(v.decode_value(), 10.0)
        v = azm.FeatureValue(name="", type=azm.FeatureType.String, value="test")
        self.assertEqual(v.decode_value(), "test")
        v = azm.FeatureValue(name="", type=azm.FeatureType.Binary, value="dGVzdA==")
        self.assertEqual(v.decode_value(), b"test")
        v = azm.FeatureValue(name="", type=azm.FeatureType.Datetime, value="2022-01-01T01:00:00+00:00")
        self.assertEqual(v.decode_value(), datetime.datetime(2022, 1, 1, 1, 0, tzinfo=datetime.timezone.utc))
        v = azm.FeatureValue(name="", type=azm.FeatureType.Filepath, value="a/b/c")
        self.assertEqual(v.decode_value(), "a/b/c")
        v = azm.FeatureValue(name="", type=azm.FeatureType.Uri, value="http://test.com")
        self.assertEqual(v.decode_value(), "http://test.com")
        v = azm.FeatureValue(name="", type=azm.FeatureType.Uri, value="http://test.com")
        self.assertEqual(v.decode_value(), "http://test.com")
        v = azm.FeatureValue(name="", type=azm.FeatureType.Uri, value="meow")
        v.type = "cat"
        self.assertRaises(exceptions_bedrock.AzulValueError, v.decode_value)



from __future__ import annotations
from pydantic import BaseModel
from typing import Any

from azul_bedrock import models_network as azm
from multiprocessing import shared_memory
from azul_runner import (
    DATA_HASH,
    FV,
    Event,
    EventData,
    EventParent,
    Feature,
    FeatureType,
    FeatureValue,
    JobResult,
    State,
    TestPlugin,
)
from tests import plugin_support as sup


SHARED_MEM_NAME = "test_plugin_grandchildren"
_add_grandchild_result_parent_sha256 = DATA_HASH(b"input stream content").hexdigest()


class GrandchildrenSharableInput(BaseModel):
    """Sharable Pydantic BaseModel to allow default dict values."""

    add_child_key: str = "foo"
    add_child_val: dict = {"rt": "rv"}
    add_grandchild_key: str = "bar"
    add_grandchild_val: dict = {"rt": "rv"}
    feat_key: str = "example_unspec"
    feat_val: Any = None


class DPGrandChildrenSharedMem(sup.DummyPluginDefaultSharedMem):
    def get_shared_memory_name(self):
        return SHARED_MEM_NAME

    def execute(self, job):
        shared_mem_val: GrandchildrenSharableInput = self._load_from_shared_memory()
        c = self._add_child(shared_mem_val.add_child_key, shared_mem_val.add_child_val)
        gc = c._add_child(shared_mem_val.add_grandchild_key, shared_mem_val.add_grandchild_val)
        gc.add_feature_values(shared_mem_val.feat_key, shared_mem_val.feat_val)


class TestPluginGrandchildHandling(sup.TestPluginSharedMem):
    """
    Tests for the handling of add_grandchild and related errors
    """

    PLUGIN_TO_TEST = sup.DummyPlugin

    @classmethod
    def setUpClass(cls) -> None:
        sup.cleanup_shared_memory(SHARED_MEM_NAME)
        cls.shared_mem = shared_memory.SharedMemory(name=SHARED_MEM_NAME, create=True, size=10240)
        return super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.shared_mem.close()
        cls.shared_mem.unlink()
        return super().tearDownClass()

    def test_add_none(self):
        # Test plugin adding grandchild feature entries that are not a valid type for a feature
        self._set_shared_memory(
            self.shared_mem,
            GrandchildrenSharableInput(
                feat_val=None,
            ),
        )
        result = self.do_execution(plugin_class=DPGrandChildrenSharedMem)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        sha256="foo",
                        relationship={"rt": "rv"},
                    ),
                    Event(
                        parent=EventParent(
                            parent=EventParent(sha256="test_entity"),
                            sha256="foo",
                            relationship={"rt": "rv"},
                        ),
                        sha256="bar",
                        relationship={"rt": "rv"},
                        features={"example_unspec": []},
                    ),
                ],
            ),
        )

        # test list of values with a None
        self._set_shared_memory(
            self.shared_mem,
            GrandchildrenSharableInput(
                feat_val=["1", "2", None],
            ),
        )
        result = self.do_execution(plugin_class=DPGrandChildrenSharedMem)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        sha256="foo",
                        relationship={"rt": "rv"},
                    ),
                    Event(
                        parent=EventParent(
                            parent=EventParent(sha256="test_entity"),
                            sha256="foo",
                            relationship={"rt": "rv"},
                        ),
                        sha256="bar",
                        relationship={"rt": "rv"},
                        features={"example_unspec": [FV("1"), FV("2")]},
                    ),
                ],
            ),
        )

    def test_add_grandchild_feature_errors(self):
        """Tests that raises the expected errors when child features are malformed"""

        # Test plugin adding a grandchild feature not defined in FEATURES
        self._set_shared_memory(
            self.shared_mem,
            GrandchildrenSharableInput(
                feat_key="feat",
                feat_val=[None],
            ),
        )
        result = self.do_execution(plugin_class=DPGrandChildrenSharedMem)
        m: str = result.state.message
        self.assertRegex(m, r"ResultError: Plugin tried to set undeclared features")
        self.assertEqual(result.state, State(State.Label.ERROR_OUTPUT, "Invalid Plugin Output", m))

        # Test plugin adding grandchild feature entries that are not a valid type for a feature
        self._set_shared_memory(
            self.shared_mem,
            GrandchildrenSharableInput(
                feat_val=555,
            ),
        )
        result = self.do_execution(plugin_class=DPGrandChildrenSharedMem)
        m: str = result.state.message
        self.assertRegex(m, "Plugin returned a value with incorrect type")
        self.assertEqual(result.state, State(State.Label.ERROR_OUTPUT, "Invalid Plugin Output", m))

        # Test plugin adding a grandchild feature item list with invalid value types in it
        self._set_shared_memory(
            self.shared_mem,
            GrandchildrenSharableInput(
                add_child_key="bar",
                add_child_val={},
                add_grandchild_key="baz",
                add_grandchild_val={},
                feat_key="example_unspec",
                feat_val=["1", "2", "3", 4],
            ),
        )
        result = self.do_execution(plugin_class=DPGrandChildrenSharedMem)
        m: str = result.state.message
        self.assertRegex(m, "Plugin returned a value with incorrect type")
        self.assertEqual(result.state, State(State.Label.ERROR_OUTPUT, "Invalid Plugin Output", m))

    def test_add_grandchild_simple_feature_value(self):
        # Test a plugin successfully adding a simple contentless grandchild and a single-value feature
        for val in sup.VALID_FEATURE_EXAMPLES:
            self._set_shared_memory(
                self.shared_mem,
                GrandchildrenSharableInput(
                    add_child_key="cid",
                    add_child_val={"reltype": "foo"},
                    add_grandchild_key="gid",
                    add_grandchild_val={"r": "t"},
                    feat_key="example_unspec",
                    feat_val=str(val),
                ),
            )
            result = self.do_execution(plugin_class=DPGrandChildrenSharedMem)

            self.assertJobResult(
                result,
                JobResult(
                    state=State(State.Label.COMPLETED),
                    events=[
                        Event(
                            parent=EventParent(sha256="test_entity"),
                            sha256="cid",
                            relationship={"reltype": "foo"},
                        ),
                        Event(
                            parent=EventParent(
                                parent=EventParent(sha256="test_entity"),
                                sha256="cid",
                                relationship={"reltype": "foo"},
                            ),
                            sha256="gid",
                            relationship={"r": "t"},
                            features={"example_unspec": [FV(str(val))]},
                        ),
                    ],
                ),
            )

        # Same test but with a FeatureValue object
        for val in sup.VALID_FEATURE_EXAMPLES:
            self._set_shared_memory(
                self.shared_mem,
                GrandchildrenSharableInput(
                    add_child_key="cid",
                    add_child_val={"reltype": "foo"},
                    add_grandchild_key="gid",
                    add_grandchild_val={},
                    feat_key="example_unspec",
                    feat_val=FeatureValue(str(val)),
                ),
            )
            result = self.do_execution(plugin_class=DPGrandChildrenSharedMem)
            self.assertJobResult(
                result,
                JobResult(
                    state=State(State.Label.COMPLETED),
                    events=[
                        Event(
                            parent=EventParent(sha256="test_entity"),
                            sha256="cid",
                            relationship={"reltype": "foo"},
                        ),
                        Event(
                            parent=EventParent(
                                parent=EventParent(sha256="test_entity"),
                                sha256="cid",
                                relationship={"reltype": "foo"},
                            ),
                            sha256="gid",
                            features={"example_unspec": [FV(str(val))]},
                        ),
                    ],
                ),
            )

    class DPAddGrandChildResults1(sup.DummyPlugin):
        """Add a child and a grandchild, and check that they are returned correctly in result['children']"""

        def execute(self, job):
            self.add_feature_values("filename", "bad.exe")
            c = self._add_child("foo", {})
            c.add_feature_values("example_unspec", ["3", "horse", "1", "2"])
            c.add_data(azm.DataLabel.TEST, {}, b"some content")
            gc = c._add_child("bar", {"rel": "extracted"})
            gc.add_feature_values("example_int", [8, 7, 6])
            gc.add_data(azm.DataLabel.CONTENT, {}, b"Grandchild binary content")

    class DPAddGrandChildResults2(sup.DummyPlugin):
        def execute(self, job):
            c1 = self._add_child("child", {})
            c2 = self._add_child("child2", {}, parent_sha256=_add_grandchild_result_parent_sha256)
            gc1 = c1._add_child("foo", {"rel": "1"})
            gc1.add_feature_values("example_unspec", ["3", "horse", "1", "2"])
            gc1.add_data(azm.DataLabel.TEXT, {}, b"some content")

            gc2 = c1.add_child_with_data({}, b"2")
            gc2.add_feature_values("example_string", ["value", "other value"])
            gc2.add_data(azm.DataLabel.ASSEMBLYLINE, {}, b"other content")
            gc2.add_data(azm.DataLabel.REPORT, {}, b"more content")

            gc3 = c2.add_child_with_data({"rel": "3"}, b"3")
            gc3.add_feature_values("example_int", [42])

    def test_add_grandchild_results(self):
        """Tests that added children return the expected result"""
        result = self.do_execution(plugin_class=self.DPAddGrandChildResults1)

        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(sha256="test_entity", features={"filename": [FV("bad.exe")]}),
                    Event(
                        parent=EventParent(sha256="test_entity", filename="bad.exe"),
                        sha256="foo",
                        data=[EventData(hash=DATA_HASH(b"some content").hexdigest(), label=azm.DataLabel.TEST)],
                        features={"example_unspec": [FV("1"), FV("2"), FV("3"), FV("horse")]},
                    ),
                    Event(
                        parent=EventParent(
                            parent=EventParent(sha256="test_entity", filename="bad.exe"),
                            sha256="foo",
                        ),
                        sha256="bar",
                        relationship={"rel": "extracted"},
                        data=[
                            EventData(
                                hash=DATA_HASH(b"Grandchild binary content").hexdigest(),
                                label=azm.DataLabel.CONTENT,
                            )
                        ],
                        features={"example_int": [FV(6), FV(7), FV(8)]},
                    ),
                ],
                data={
                    DATA_HASH(b"some content").hexdigest(): b"some content",
                    DATA_HASH(b"Grandchild binary content").hexdigest(): b"Grandchild binary content",
                },
            ),
            inspect_data=True,
        )

        # Add multiple grandchildren, one under a child in a stream-specific result
        some_content = DATA_HASH(b"some content").hexdigest()
        other_content = DATA_HASH(b"other content").hexdigest()
        more_content = DATA_HASH(b"more content").hexdigest()
        content_2 = DATA_HASH(b"2").hexdigest()
        content_3 = DATA_HASH(b"3").hexdigest()
        result = self.do_execution(
            plugin_class=self.DPAddGrandChildResults2, data_in=[(azm.DataLabel.TEST, b"input stream content")]
        )
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        sha256="child",
                    ),
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        parent_sha256=_add_grandchild_result_parent_sha256,
                        sha256="child2",
                    ),
                    Event(
                        parent=EventParent(
                            parent=EventParent(sha256="test_entity"),
                            sha256="child",
                        ),
                        sha256="foo",
                        relationship={"rel": "1"},
                        data=[EventData(hash=some_content, label=azm.DataLabel.TEST)],
                        features={"example_unspec": [FV("1"), FV("2"), FV("3"), FV("horse")]},
                    ),
                    Event(
                        parent=EventParent(
                            parent=EventParent(sha256="test_entity"),
                            sha256="child",
                        ),
                        sha256=content_2,
                        data=[
                            EventData(hash=content_2, label=azm.DataLabel.CONTENT),
                            EventData(hash=other_content, label=azm.DataLabel.TEXT),
                            EventData(hash=more_content, label=azm.DataLabel.TEST),
                        ],
                        features={"example_string": [FV("other value"), FV("value")]},
                    ),
                    Event(
                        parent=EventParent(
                            parent=EventParent(sha256="test_entity"),
                            parent_sha256=_add_grandchild_result_parent_sha256,
                            sha256="child2",
                        ),
                        sha256=content_3,
                        relationship={"rel": "3"},
                        data=[EventData(hash=content_3, label=azm.DataLabel.CONTENT)],
                        features={"example_int": [FV(42)]},
                    ),
                ],
                data={
                    some_content: b"some content",
                    content_2: b"2",
                    other_content: b"other content",
                    more_content: b"more content",
                    content_3: b"3",
                },
            ),
            inspect_data=True,
        )

    class DPAddFilenames(sup.DummyPlugin):
        FEATURES = [
            Feature("filename", "Example string feature", type=FeatureType.String),
        ]

        def execute(self, job):
            c = self._add_child("cid", {"reltype": "dummy"})
            c.add_feature_values("filename", "alpha.3")
            c.add_feature_values("filename", "alpha.1")
            c2 = c._add_child("cid2", {"reltype": "dummy2"})

    def test_add_filenames(self):
        """Test that filename feature values for EventParent() are sorted before applying - for consistency."""
        result = self.do_execution(plugin_class=self.DPAddFilenames)
        # Should pick up the plugin's default entity type since it's set to None in the add_child call
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        sha256="cid",
                        relationship={"reltype": "dummy"},
                        features={"filename": [FV("alpha.1"), FV("alpha.3")]},
                    ),
                    Event(
                        parent=EventParent(
                            parent=EventParent(sha256="test_entity"),
                            sha256="cid",
                            relationship={"reltype": "dummy"},
                            filename="alpha.1",
                        ),
                        sha256="cid2",
                        relationship={"reltype": "dummy2"},
                    ),
                ],
            ),
        )



"""Defines a file-like interface to S3-compatible storage.

Used by plugins to access event binary data.
"""

import hashlib
import io
import logging
import re
import typing
from typing import Optional

import httpx
from azul_bedrock import dispatcher
from azul_bedrock import models_network as azm
from azul_bedrock.exceptions_bedrock import DispatcherApiException
from bitarray import bitarray

from . import storage_spooled

# Matches the Content-Range: header in server responses
RANGE_RE = re.compile(r"bytes ([0-9]+)-([0-9]+)/([0-9]+|\*)$")

logger = logging.getLogger(__name__)

# Declare the hash algorithm to be used for data stream identification
DATA_HASH_NAME = "sha256"
DATA_HASH = hashlib.sha256


def calc_stream_hash(binaryio: typing.BinaryIO, hash_type=DATA_HASH) -> str:
    """Calculate the hexdigest of a binary file."""
    binaryio.seek(0)
    h = hash_type()
    while True:
        # read 1MB at a time
        buf = binaryio.read(2**20)
        if not buf:
            break
        h.update(buf)
    binaryio.seek(0)
    return h.hexdigest()


class StorageError(Exception):  # noqa: D204
    """StorageException.

    The storage system has failed in some way. This indicates a system fault such as a server timeout;
     user errors (eg seek to negative offset, read of closed file) should raise the usual exceptions.
    """

    pass


class ProxyFileNotFoundError(StorageError, FileNotFoundError):
    """Storage Proxy File backing not found."""


class StorageProxyFile(io.RawIOBase):
    """StorageProxyFile class.

    Access to storage/data proxy API via a file-like object.
    This is intended to avoid the need for fetching entire files/data streams if the plugin only needs part of it.

    Data will be fetched in multiples of chunk_size and cached for future httpx. The chunk_size should be relatively
     large in order to minimise number of GETs, especially when reading small pieces of data one after the other or
     spread out through a particular range; however, you may wish to reduce it if your network is slow and you
     know you will only need small, randomly-distributed parts of the file.
    NOTE: There is no support for resources that can change (eg if-modified-since etc), as this is intended to be used
     with immutable data objects identified by content hash.
    """

    tags: dict[str, str]  # Tags for the data stream - hashes, MIME types, whatever you want

    _url: str
    _session: httpx.Client  # a requests session allows us to re-use the same connection
    _offset: int
    _size: Optional[int]  # None if size is currently unknown.
    _chunk_size: int  # Chunk size for content caching
    _chunks: bitarray  # Bitfield indicating which chunks have data (may be zero-len)
    _content: storage_spooled.SpooledNamedTemporaryFile  # cached content data

    # If the last chunk is less than this fraction of a full chunk, fetch it and the second-last chunk together.
    MIN_LAST_CHUNK_SIZE: float = 0.66

    def __init__(
        self,
        source: str,
        label: azm.DataLabel,
        hash: str,
        *,
        dp: dispatcher.DispatcherAPI = None,
        request_timeout: int = 10,
        # retrieve in 1mb chunks
        chunk_size: int = 1_048_576,
        # cache ~20 mb in memory before moving to disk
        mem_cache_limit: int = 20_971_520,
        init_data: bytes | None = None,
        expected_size: int | None = None,
        file_info: azm.Datastream | None = None,
        allow_unbounded_read: bool = True,
    ):
        """Init.

        :param url: URL of the file to proxy
        :param request_timeout: Timeout for HTTP connections and requests
        :param chunk_size: Size of data chunks for caching; always request full chunks for every read.
                           This should be relatively large, to avoid excessive numbers of GET requests (default 256k).
        :param mem_cache_limit: size limit for memory caching before falling back to a disk file (default 64M).
        :param init_data: Initialise the instance with existing byte data. This must be the full content of the file;
                          if this is set, url is ignored and no HTTP requests are made.
        :param expected_size: Initialises our data structures for the expected size of the file. This saves us from
                              having to fetch the first chunk of file to determine the size on first access.
        """
        self._source = source
        self._label = label
        self._hash = hash

        self._timeout = request_timeout
        self._content = storage_spooled.SpooledNamedTemporaryFile(max_size=mem_cache_limit)
        self._offset = 0
        self._dispatcher = dp
        self._allow_unbounded_read = allow_unbounded_read
        self.file_info = file_info
        if isinstance(init_data, bytes):
            logger.debug("Init StorageProxyFile with init_data (%d bytes), hash=%s" % (len(init_data), self._hash))
            # noinspection PyTypeChecker
            self._session = None  # Unset because full data is already provided.
            self._content.write(init_data)
            self._chunk_size = self._size = len(init_data)
            self._chunks = bitarray([True])
            if expected_size is not None and expected_size != len(init_data):
                raise ValueError("expected_size does not match provided init_data")
        else:
            logger.debug(
                "Init StorageProxyFile with hash=%s%s"
                % (self._hash, (", expected size %d" % expected_size) if expected_size else ", size unknown")
            )
            self._session = httpx.Client(timeout=5.0)
            self._chunk_size = chunk_size
            self._size = expected_size
            self._chunks = bitarray()
            if self._size is not None:
                self._recalc_chunks(self._size)

    def __setstate__(self, state: dict):
        """Used to set the state of the object after pickling."""
        self._source = state.get("_source", None)
        self._label = state.get("_label", None)
        self._hash = state.get("_hash", None)
        self._timeout = state.get("_timeout", None)
        self._content = storage_spooled.SpooledNamedTemporaryFile(max_size=state.get("mem_cache_limit", 20_971_520))
        self._content.write(state.get("_content", b""))
        self._offset = state.get("_offset", None)
        self._dispatcher = state.get("_dispatcher", None)
        self._allow_unbounded_read = state.get("_allow_unbounded_read", None)
        self.file_info = state.get("file_info", None)
        self._session = state.get("_session", None)
        self._chunk_size = state.get("_chunk_size", None)
        self._chunks = state.get("_chunks", None)
        self._size = state.get("_size", None)

    def __getstate__(self):
        """Used to save the state of an object prior to pickling."""
        # Ensure full content is retrieved.
        self._retrieve(-1)
        return {
            "_source": self._source,
            "_label": self._label,
            "_hash": self._hash,
            "_timeout": self._timeout,
            "mem_cache_limit": self._content._max_size,
            # Need to read content in because buffered random can't be read in.
            "_content": self._content.read(),
            "_offset": self._offset,
            "_dispatcher": self._dispatcher,
            "_allow_unbounded_read": self._allow_unbounded_read,
            "file_info": self.file_info,
            "_session": self._session,
            "_chunk_size": self._chunk_size,
            "_chunks": self._chunks,
            "_size": self._size,
        }

    def get_tempfile(self) -> storage_spooled.SpooledNamedTemporaryFile:
        """Read all data and return the temporary file.

        This is useful for third party libraries that take a file object, as it
        can reduces memory usage vs a bytearray.
        """
        # go to beginning of file
        self.seek(0)
        # read all bytes into temporary file
        self._retrieve(-1)
        return self._content

    def get_filepath(self) -> str:
        """Read all data and return a filepath.

        This is useful for third party libraries that take a file path, as it
        can reduces memory usage vs a bytearray.
        """
        # go to beginning of file
        self.seek(0)
        # read all bytes into temporary file
        self._retrieve(-1)
        # force write to disk if under size threshold
        self._content.rollover()
        return self._content.name

    def get_hash(self) -> str:
        """Return hash for identifying this data."""
        return self.file_info.sha256

    def _recalc_chunks(self, size: int):
        """(Re)initialises the chunks array for our file based on a given file size."""
        logger.debug("Recalculating chunks (size %d) for %s" % (size, self._hash))
        new_len = 1 + (size - 1) // self._chunk_size
        if len(self._chunks) > new_len:
            raise RuntimeError("Cannot shrink existing StorageProxyFile - unexpected change in size")
        b = bitarray(new_len - len(self._chunks))
        b.setall(False)
        self._chunks.extend(b)

    def close(self) -> None:
        """Close and delete the underlying file or network connection."""
        if self.closed:
            return
        if self._session:
            self._session.close()
        self._content.close()
        del self._content
        del self._chunks
        super(StorageProxyFile, self).close()

    def _assert_size(self, size: int) -> None:
        """assert_size; asserts that the object's size hasn't changed."""
        logger.debug("Assert size: %s == %s" % (self._size, size))
        if self._size != size:
            raise AssertionError("Object size mismatch for %s" % self._hash)

    def _assert_content_len(self, resp: httpx.Response) -> None:
        """assert_content_len. Verifies that the content's length matches the Content-Length header."""
        if "Content-Length" in resp.headers and int(resp.headers["Content-Length"]) != len(resp.content):
            raise StorageError("Content length mismatch for %s" % self._hash)

    def _make_request(self, start_pos: Optional[int], end_pos: Optional[int]) -> httpx.Response:
        """make_request. Wraps the GET request and handles errors, raising the appropriate exceptions."""

        def s_fmt(v):
            return v if v is not None else ""

        try:
            logger.error(f"Requesting bytes={s_fmt(start_pos)}-{s_fmt(end_pos)} of {s_fmt(self._hash)}")
            resp = self._dispatcher.get_binary(
                source=self._source, label=self._label, sha256=self._hash, start_pos=start_pos, end_pos=end_pos
            )
        except DispatcherApiException as e:
            if e.status_code == 404:
                raise ProxyFileNotFoundError(2, "Got 404 requesting %s" % self._hash) from e
            elif e.status_code == 416:  # Range not satisfiable; return this to the caller for processing
                resp = e.response
            else:
                raise StorageError("Got %s requesting %s" % (e.status_code, self._hash)) from e
        if resp.headers["Content-Type"] != "application/octet-stream":
            raise StorageError(
                "Server provided %s instead of octet-stream for %s" % (resp.headers["Content-Type"], self._hash)
            )
        return resp

    def _handle_full_content(self, resp: httpx.Response):
        """When called, takes a response containing the full file and sets self._content/self._chunks appropriately."""
        self._assert_content_len(resp)
        if self._size:
            self._assert_size(len(resp.content))
        if "Content-Range" in resp.headers:
            raise StorageError("Content-Range header unexpectedly present in 200 response for %s" % self._hash)
        # Handle the file as a single 'chunk', as no further fetches are needed.
        self._chunk_size = self._size = len(resp.content)
        self._chunks = bitarray([True])
        self._content.seek(0)
        self._content.write(resp.content)

    def _parse_partial_response(self, resp: httpx.Response) -> tuple[str, str, str]:
        """Return (request start, end, total file size) if response is valid."""
        self._assert_content_len(resp)
        if "Content-Range" not in resp.headers:
            # RFC specifies 'server MUST generate a Content-Range header'; if missing, server is broken.
            raise StorageError("Server failed to provide Content-Range header for %s" % self._hash)
        if "bytes */" in resp.headers["Content-Range"]:
            # 'unsatisfied range' response header - RFC implies this is only valid with 416 code
            raise StorageError("Server returned 206 with 'Content-Range: bytes */...' for %s" % self._hash)
        hits = RANGE_RE.match(resp.headers["Content-Range"])
        if not hits:
            raise StorageError(
                "Could not parse 'Content-Range: %s' (requesting %s)" % (resp.headers["Content-Range"], self._hash)
            )
        s, e, t = hits.groups()  # Ensures that there are exactly 3 hits to unpack
        return s, e, t

    def _request_chunks(self, start_chunk: Optional[int], end_chunk: Optional[int]) -> None:
        """Handles requests for file data.

        end_chunk is exclusive, like slice notation; fetch up to end_chunk - 1.
        This method fetches the data, updates self._chunks, and stores the data to self._content.

        Special cases:
            end_chunk = None reads all remaining data, the same as read(-1).
            start_chunk = None reads the last chunk of file (used to definitively determine size; end_chunk ignored)
        """
        # Variable naming:
        #  start_chunk, end_chunk - the chunks to be fetched
        #  req_start, req_end - the byte ranges requested from the server
        #  resp_start, resp_end, resp_total - the range the server actually sent (and total size, if provided)

        if self.closed:
            raise ValueError("_request_chunks called on closed file")
        if start_chunk is not None and start_chunk < 0:
            raise ValueError("start_chunk cannot be negative")
        logger.debug("Get chunks %s...%s for %s" % (start_chunk, end_chunk, self._hash))
        if end_chunk is not None and start_chunk is not None:
            if end_chunk < start_chunk:
                raise ValueError("end_chunk must be >= start_chunk")
            if end_chunk == start_chunk:
                return  # No-op
            if self._size is not None and start_chunk > (self._size - 1) // self._chunk_size:
                return

        if start_chunk is None:
            req_start = None
            req_end = self._chunk_size  # Note: not negative, because it will be subbed into '-%s'
        else:
            req_start = start_chunk * self._chunk_size
            if end_chunk is None:
                req_end = None
            else:
                # end_chunk is exclusive, so need to request up to the byte just before that chunk
                req_end = end_chunk * self._chunk_size - 1
                if self._size is not None and req_end >= self._size:
                    req_end = self._size - 1

        resp = self._make_request(req_start, req_end)
        if resp.status_code == 416:
            # Range not satisfiable -- must be past EOF
            if req_start is None:
                raise StorageError("Server returned 416 for a suffix range request - broken server?")
            if "Content-Range" in resp.headers:
                # server 'SHOULD' generate a Content-Range header with current length in a 416
                cr: str = resp.headers["Content-Range"]
                if not cr.startswith("bytes */"):
                    raise StorageError(
                        "Server returned unexpected Content-Range header for 416 ('%s')"
                        % resp.headers["Content-Range"]
                    )
                else:
                    if cr == "bytes */*":
                        logger.warning("Server failed to return content size in 416 Content-Range header")
                    else:
                        self._size = int(cr[8:])
            return  # Nothing further to do
        elif resp.status_code == 200:
            # Full content returned, even if it wasn't requested.
            self._handle_full_content(resp)
        elif resp.status_code == 206:
            # Partial Content
            resp_start, resp_end, resp_total = self._parse_partial_response(resp)
            if resp_total != "*":
                if self._size is None:
                    self._size = int(resp_total)
                    self._recalc_chunks(self._size)
                    if req_start is None:
                        # This was a request for the last chunk of file; adjust expected response values
                        req_start = max(self._size - self._chunk_size, 0)
                        req_end = self._size - 1
                        end_chunk = 1 + req_end // self._chunk_size  # end_chunk is exclusive
                        start_chunk = end_chunk - 1
                    elif req_end is None or req_end >= self._size:
                        # Adjust the expected end byte to avoid raising an exception below
                        req_end = self._size - 1
                else:
                    self._assert_size(int(resp_total))
            elif req_start is None:
                # Get 'the last <chunk_size> bytes of the file'; calculate length from the response
                req_end = int(resp_end)
                self._size = req_end + 1
                req_start = max(self._size - self._chunk_size, 0)
                end_chunk = 1 + req_end // self._chunk_size  # end_chunk is exclusive
                start_chunk = end_chunk - 1  # Only asked for one chunk
                self._recalc_chunks(self._size)
            elif req_end is None:
                # Get "all the rest of the file", so assume the end of the response is the total size
                req_end = int(resp_end)
                end_chunk = 1 + req_end // self._chunk_size  # end_chunk is exclusive
                if self._size is None:
                    self._size = req_end + 1
                    self._recalc_chunks(self._size)
                else:
                    self._assert_size(req_end + 1)

            if (int(resp_start), int(resp_end)) != (req_start, req_end):
                if self._size is None and int(resp_start) == req_start:
                    # Got less than the total requested size; assume it's because of EOF
                    self._size = int(resp_end) + 1
                    self._recalc_chunks(self._size)
                else:
                    raise StorageError(
                        "Server returned unrequested range: req %s != resp %s (requesting %s)"
                        % (resp.request.headers["Range"], resp.headers["Content-Range"], self._hash)
                    )
            self._assert_content_len(resp)
            # Add data to cache
            self._content.seek(int(resp_start))
            self._content.write(resp.content)
            if self._size is None and end_chunk > len(self._chunks):
                # Don't yet know the true size, but need to extend our chunks to mark the data acquired so far.
                self._recalc_chunks(int(resp_end) + 1)
            self._chunks[start_chunk:end_chunk] = True
        else:
            # If this point is reached, it's not an error but also not '200 OK' or '206 Partial Content'
            raise StorageError("Unexpected response code %s while requesting %s" % (resp.status_code, self._hash))

    def _retrieve(self, size: int):
        """Retrieve bytes from remote storage into the temporary file."""
        if not isinstance(size, int):
            raise TypeError("read length must be an int")
        if self.closed:
            raise ValueError("read of closed file")
        if size < -1:
            raise ValueError("read length must be positive or -1")
        logger.debug("Read %s@%s for %s" % (size, self._offset, self._hash))
        if self._size is not None:
            # Limit size to the remaining size of data in the file
            if size == -1:
                size = max(self._size - self._offset, 0)
            else:
                size = max(min(self._size - self._offset, size), 0)
        if size == 0:
            logger.debug("Short-circuiting zero-size read")
            return b""

        # Calculate start and end chunks. end_chunk is exclusive, like in slice notation.
        start_chunk = self._offset // self._chunk_size
        if size == -1:
            # This implies that self._size is None, else size would have been calculated above
            end_chunk = -1
        else:
            end_chunk = 1 + (self._offset + size - 1) // self._chunk_size

        if self._size is not None:
            # Special-case the last chunk together with the second-last if it's small
            if len(self._chunks) > 1 and (
                (self._size - 1) % self._chunk_size + 1 < self.MIN_LAST_CHUNK_SIZE * self._chunk_size
            ):
                if end_chunk == len(self._chunks) - 1:
                    end_chunk += 1
                elif start_chunk == len(self._chunks) - 1:
                    start_chunk -= 1

        # Check for any missing data chunks in the desired range that will need to be fetched
        while not all(self._chunks[start_chunk:end_chunk]):
            # Find first missing chunk
            next_req_start = self._chunks.index(False, start_chunk, end_chunk)
            # Find the first cached chunk after that (if any) in the requested range,
            #  so the same data isn't re-requested.
            if any(self._chunks[next_req_start:end_chunk]):
                next_req_end = self._chunks.index(True, next_req_start, end_chunk)
            else:
                next_req_end = end_chunk
            self._request_chunks(next_req_start, None if next_req_end == -1 else next_req_end)
        if self._size is None and (end_chunk == -1 or len(self._chunks) < end_chunk):
            # This should only happen if NO chunks have been acquired yet, or if some chunks have been fetched
            #  but the server hasn't told us the total file size.
            self._request_chunks(max(start_chunk, len(self._chunks)), None if end_chunk == -1 else end_chunk)

        # reset seek position in case it was changed during collect
        self._content.seek(self._offset)

    def read(self, size: int = -1) -> bytes:
        """Read the provided number of bytes or an unbounded read of the whole file."""
        if size < 0 and not self._allow_unbounded_read:
            raise ValueError("unbounded/full read() calls are disabled due to memory usage")
        self._retrieve(size)
        # Should now have all the required chunks
        rv = self._content.read(size)
        self._offset += len(rv)
        return rv

    def seekable(self) -> bool:
        """Is the file seekable - Always True."""
        return True

    def readable(self) -> bool:
        """Is the file is readable - Always True."""
        return True

    def writable(self) -> bool:
        """Is the file is writable - Always False."""
        return False

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        """Seek to a specified offset within the file."""
        if not isinstance(offset, int):
            raise TypeError("offset must be an int")
        if self.closed:
            raise ValueError("seek of closed file")

        if whence == io.SEEK_SET:
            self._offset = offset
        elif whence == io.SEEK_CUR:
            self._offset += offset
        elif whence == io.SEEK_END:
            if self._size is None:
                # Request the last full chunk of the file. It's the only way to be 100% sure of finding the total size.
                self._request_chunks(start_chunk=None, end_chunk=-1)
            self._offset = self._size + offset
        else:
            raise ValueError("whence value %s unsupported" % whence)
        if self._offset < 0:
            raise OSError(22, "Invalid argument")  # Duplicate behaviour of regular file
        # No errors for offsets beyond end of file; they just return nothing from read()

        return self._offset

    def tell(self) -> int:
        """Get the current offset without moving it."""
        if self.closed:
            raise ValueError("I/O operation on closed file")
        return self._offset



from __future__ import annotations

from dataclasses import dataclass
import tempfile
from typing import Any

from azul_bedrock import models_network as azm

from azul_runner import (
    DATA_HASH,
    FV,
    Event,
    EventData,
    EventParent,
    FeatureValue,
    JobResult,
    State,
)
from multiprocessing import shared_memory
from . import plugin_support as sup

SHARED_MEM_NAME = "test_plugin_children_shared_mem"


@dataclass
class SharableInput:
    child_key: str
    child_val: dict
    feat_key: str
    feat_val: Any


class DpGenericAddChildAndFeature(sup.DummyPluginDefaultSharedMem):
    def get_shared_memory_name(self):
        return SHARED_MEM_NAME

    def execute(self, job):
        shared_mem_val: SharableInput = self._load_from_shared_memory()
        c = self._add_child(shared_mem_val.child_key, shared_mem_val.child_val)
        c.add_feature_values(shared_mem_val.feat_key, shared_mem_val.feat_val)


class TestPluginChildHandling(sup.TestPluginSharedMem):
    """
    Tests for the handling of add_child, add_child_data, and adding children in run_once
    """

    PLUGIN_TO_TEST = sup.DummyPlugin

    @classmethod
    def setUpClass(cls) -> None:
        sup.cleanup_shared_memory(SHARED_MEM_NAME)
        cls.shared_mem = shared_memory.SharedMemory(name=SHARED_MEM_NAME, create=True, size=10240)
        return super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.shared_mem.close()
        cls.shared_mem.unlink()
        return super().tearDownClass()

    def test_add_none(self):
        # Test plugin adding child feature entries that are not a valid type for a feature
        self._set_shared_memory(
            self.shared_mem,
            SharableInput(child_key="foo", child_val={"rt": "rv"}, feat_key="example_unspec", feat_val=None),
        )
        result = self.do_execution(
            plugin_class=DpGenericAddChildAndFeature,
        )
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        sha256="foo",
                        relationship={"rt": "rv"},
                        features={"example_unspec": []},
                    )
                ],
            ),
        )

        self._set_shared_memory(
            self.shared_mem,
            SharableInput(
                child_key="foo",
                child_val={"rt": "rv"},
                feat_key="example_unspec",
                feat_val=["1", "2", None],
            ),
        )
        result = self.do_execution(
            plugin_class=DpGenericAddChildAndFeature,
        )
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        sha256="foo",
                        relationship={"rt": "rv"},
                        features={"example_unspec": [FV("1"), FV("2")]},
                    )
                ],
            ),
        )

    def test_add_child_feature_errors(self):
        """Tests that raises the expected errors when child features are malformed"""
        self._set_shared_memory(
            self.shared_mem, SharableInput(child_key="foo", child_val={"rt": "rv"}, feat_key="feat", feat_val=[None])
        )
        result = self.do_execution(
            plugin_class=DpGenericAddChildAndFeature,
        )
        m: str = result.state.message
        self.assertRegex(m, r"ResultError: Plugin tried to set undeclared features")
        self.assertEqual(result.state, State(State.Label.ERROR_OUTPUT, "Invalid Plugin Output", m))

        self._set_shared_memory(
            self.shared_mem,
            SharableInput(child_key="foo", child_val={"rt": "rv"}, feat_key="example_unspec", feat_val=555),
        )
        result = self.do_execution(
            plugin_class=DpGenericAddChildAndFeature,
        )
        m: str = result.state.message
        self.assertRegex(m, "Plugin returned a value with incorrect type")
        self.assertEqual(result.state, State(State.Label.ERROR_OUTPUT, "Invalid Plugin Output", m))

        self._set_shared_memory(
            self.shared_mem,
            SharableInput(
                child_key="foo", child_val={"rt": "rv"}, feat_key="example_unspec", feat_val=["1", "2", "3", 4]
            ),
        )
        result = self.do_execution(
            plugin_class=DpGenericAddChildAndFeature,
        )
        m: str = result.state.message
        self.assertRegex(m, "Plugin returned a value with incorrect type")
        self.assertEqual(result.state, State(State.Label.ERROR_OUTPUT, "Invalid Plugin Output", m))

        self._set_shared_memory(
            self.shared_mem,
            SharableInput(
                child_key="ID",
                child_val={"action": "blah"},
                feat_key="example_string",
                feat_val=[FeatureValue("a", label="foo"), FeatureValue("b", label="bar")],
            ),
        )
        result = self.do_execution(
            plugin_class=DpGenericAddChildAndFeature,
        )
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        sha256="ID",
                        relationship={"action": "blah"},
                        features={"example_string": [FV("a", label="foo"), FV("b", label="bar")]},
                    )
                ],
            ),
        )

    def test_add_child_simple_feature_value(self):
        # Test a plugin successfully adding a simple contentless child and encapsulating single-value feature
        for val in sup.VALID_FEATURE_EXAMPLES:
            self._set_shared_memory(
                self.shared_mem,
                SharableInput(
                    child_key="cid", child_val={"reltype": "dummy"}, feat_key="example_unspec", feat_val=str(val)
                ),
            )
            result = self.do_execution(
                plugin_class=DpGenericAddChildAndFeature,
            )
            # Should pick up the plugin's default entity type since it's set to None in the add_child call
            self.assertJobResult(
                result,
                JobResult(
                    state=State(State.Label.COMPLETED),
                    events=[
                        Event(
                            parent=EventParent(sha256="test_entity"),
                            sha256="cid",
                            relationship={"reltype": "dummy"},
                            features={"example_unspec": [FV(str(val))]},
                        )
                    ],
                ),
            )
        # Same test but with a FeatureValue object
        for val in sup.VALID_FEATURE_EXAMPLES:
            self._set_shared_memory(
                self.shared_mem,
                SharableInput(
                    child_key="cid", child_val={"example": "foo"}, feat_key="example_unspec", feat_val=FV(str(val))
                ),
            )
            result = self.do_execution(
                plugin_class=DpGenericAddChildAndFeature,
            )
            self.assertJobResult(
                result,
                JobResult(
                    state=State(State.Label.COMPLETED),
                    events=[
                        Event(
                            parent=EventParent(sha256="test_entity"),
                            sha256="cid",
                            relationship={"example": "foo"},
                            features={"example_unspec": [FV(str(val))]},
                        )
                    ],
                ),
            )

    class DPAddChildResults1(sup.DummyPlugin):
        # Check the expected child and features are returned
        def execute(self, job):
            c = self._add_child("foo", {})
            c.add_feature_values("example_unspec", ["3", "horse", "1", "2"])
            c.add_data(azm.DataLabel.TEST, {}, b"some content")

    class DPAddChildResults2(sup.DummyPlugin):
        # Ensure that children are not returned when the result is 'not complete'
        def execute(self, *args):
            c = self._add_child("foo", {"rel": "1"})
            c.add_feature_values("example_unspec", ["3", "horse", "1", "2"])
            c.add_data(azm.DataLabel.TEST, {}, b"some content")
            return State(State.Label.ERROR_INPUT, "Generic Error")

    class DPAddChildResults3(sup.DummyPlugin):
        # Add multiple children, and additional data for some
        def execute(self, *args):
            self.add_feature_values("example_unspec", ["1337"])
            self.add_feature_values("filename", ["bad.exe"])
            c = self.add_child_with_data({"rel": "1"}, b"some content")
            c.add_feature_values("example_unspec", ["3", "horse", "1", "2"])
            c2 = self.add_child_with_data({"rel": 2}, b"other content")
            c2.add_feature_values("example_string", ["value", "other value"])
            c2.add_data(azm.DataLabel.REPORT, {"language": "value2"}, b"more content")
            # duplicate of earlier child but with a different feature
            c3 = self.add_child_with_data({"rel": "1"}, b"some content")
            c3.add_feature_values("example_unspec", ["some new value"])
            c3.add_data(azm.DataLabel.TEST, {"language": "value"}, b"some more content")

    def test_add_child_results(self):
        """Tests that added children return the expected result"""
        result = self.do_execution(plugin_class=self.DPAddChildResults1)
        data_hash = DATA_HASH(b"some content").hexdigest()
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        sha256="foo",
                        data=[EventData(hash=data_hash, label=azm.DataLabel.TEST)],
                        features={"example_unspec": [FV("1"), FV("2"), FV("3"), FV("horse")]},
                    )
                ],
                data={data_hash: b"some content"},
            ),
            inspect_data=True,
        )

        result = self.do_execution(plugin_class=self.DPAddChildResults2)
        self.assertJobResult(result, JobResult(state=State(State.Label.ERROR_INPUT, failure_name="Generic Error")))

        some_content = DATA_HASH(b"some content").hexdigest()
        other_content = DATA_HASH(b"other content").hexdigest()
        more_content = DATA_HASH(b"more content").hexdigest()
        some_more_content = DATA_HASH(b"some more content").hexdigest()
        result = self.do_execution(plugin_class=self.DPAddChildResults3)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="test_entity",
                        features={"example_unspec": [FV("1337")], "filename": [FV("bad.exe")]},
                    ),
                    Event(
                        parent=EventParent(sha256="test_entity", filename="bad.exe"),
                        sha256=some_content,
                        relationship={"rel": "1"},
                        data=[
                            EventData(hash=some_content, label=azm.DataLabel.CONTENT),
                            EventData(hash=some_more_content, label=azm.DataLabel.TEST, language="value"),
                        ],
                        features={"example_unspec": [FV("1"), FV("2"), FV("3"), FV("horse"), FV("some new value")]},
                    ),
                    Event(
                        parent=EventParent(sha256="test_entity", filename="bad.exe"),
                        sha256=other_content,
                        relationship={"rel": 2},
                        data=[
                            EventData(hash=other_content, label=azm.DataLabel.CONTENT),
                            EventData(hash=more_content, label=azm.DataLabel.TEST, language="value2"),
                        ],
                        features={"example_string": [FV("other value"), FV("value")]},
                    ),
                ],
                data={
                    some_content: b"some content",
                    other_content: b"other content",
                    more_content: b"more content",
                    some_more_content: b"some more content",
                },
            ),
            inspect_data=True,
        )

    class DPAddEmpty(sup.DummyPlugin):
        def execute(self, job):
            c = self.add_child_with_data({"rt": "rv"}, b"")
            c.add_feature_values("example_string", [None])

    def test_add_empty(self):
        """Tests that raises the expected errors when empty children are added."""

        # Test plugin adding a child feature not defined in FEATURES
        result = self.do_execution(plugin_class=self.DPAddEmpty)
        m: str = result.state.message
        self.assertRegex(m, r"tried to add data file with 0 bytes")
        self.assertEqual(result.state, State(State.Label.ERROR_EXCEPTION, "ValueError", m))

    class DPAddChildResultsFile(sup.DummyPlugin):
        def execute(self, job):
            with tempfile.TemporaryFile("r+b") as tf1, tempfile.TemporaryFile("r+b") as tf2:
                # 100mb of 'a'/'b'
                for _ in range(100):
                    tf1.write(b"a" * 1_000_000)
                    tf2.write(b"b" * 1_000_000)
                c = self.add_child_with_data_file({"thing": "yeah"}, tf1)
                c.add_data_file(azm.DataLabel.TEST, {"language": "english"}, tf2)

    def test_add_child_results_file(self):
        """Tests that added children return the expected result"""

        # Check the expected child and features are returned
        result = self.do_execution(plugin_class=self.DPAddChildResultsFile)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        sha256="83d30385a4a11980275dc23de3fb49ff37b906cc841efa048a96c62d90ff3b5f",
                        relationship={"thing": "yeah"},
                        data=[
                            EventData(
                                hash="83d30385a4a11980275dc23de3fb49ff37b906cc841efa048a96c62d90ff3b5f",
                                label=azm.DataLabel.CONTENT,
                            ),
                            EventData(
                                hash="1854ac434080022f8c7addd0d7d79199ad38a8c551b950f3b37a76aee3c08da7",
                                label=azm.DataLabel.TEST,
                                language="english",
                            ),
                        ],
                    )
                ],
                data={
                    "83d30385a4a11980275dc23de3fb49ff37b906cc841efa048a96c62d90ff3b5f": b"",
                    "1854ac434080022f8c7addd0d7d79199ad38a8c551b950f3b37a76aee3c08da7": b"",
                },
            ),
        )



from __future__ import annotations

from dataclasses import dataclass
import datetime
import logging
from multiprocessing import shared_memory
import pickle
import time
from typing import Any, Callable

import httpx

from azul_runner import Feature, FeatureType, Job, Plugin, add_settings
from azul_runner.models import FV, State
from azul_bedrock import models_network as azm
from . import mock_dispatcher as md
from azul_runner.test_utils.test_template import TestPlugin

# Example instances of every feature type in azul.runner.models.VALID_FEATURE_TYPES
VALID_FEATURE_EXAMPLES = (101, 55.5, "string", b"BYTES1011", datetime.datetime.now(datetime.timezone.utc))


def setup_mock_dispatcher() -> tuple[md.MockDispatcher, str]:
    """Setup a mock dispatcher server and return the mock_server and the server url."""
    mock_server = md.RunnerMockDispatcher()
    mock_server.start()
    total_sleep = 0
    while not mock_server.is_alive() and total_sleep < 20:
        time.sleep(0.5)  # Wait for server to start
        total_sleep += 1
    server = "http://%s:%s" % (mock_server.host, mock_server.port)
    # Wait for server to be ready to respond
    tries = 0
    while True:
        time.sleep(0.2)
        tries += 1
        try:
            resp = httpx.get(server + "/mock/get_var/fetch_count")
            if resp.is_error:
                raise httpx.ConnectError(f"Failing to connect to dispatcher on {server}/mock/get_var/fetch_count")
            break  # Exit loop if successful
        except (httpx.TimeoutException, httpx.ConnectError):
            if tries > 20:  # Time out after about 4 seconds
                mock_server.stop()
                raise RuntimeError("Timed out waiting for mock server to be ready")
    return mock_server, server


class DummyLogHandler(logging.Handler):
    """Implements a handler that simply records any log messages generated in `self.logs`."""

    logs: list[str]

    def __init__(self):
        super().__init__()
        self.logs = []

    def emit(self, record: logging.LogRecord) -> None:
        self.logs.append("%s: %s" % (record.levelname, record.getMessage()))


# ################################################
# Dummy plugin definitions for use by test cases
# ################################################


class DummyPluginMinimum(Plugin):
    """
    Test class that will register successfully but has no execute method.
    """

    SECURITY = None
    VERSION = "none"
    FEATURES = []


# noinspection PyAbstractClass
class DummyPluginNoExecute(Plugin):
    """
    Test class that will register successfully but has no execute method.
    """

    SETTINGS = add_settings(request_retry_count=0)  # Don't retry failed requests when testing
    SECURITY = None
    VERSION = "none"
    FEATURES = []


class DummyPluginNotReady(DummyPluginNoExecute):
    """
    Test class that will always report the plugin is not ready for jobs and raise if given any.
    """

    def is_ready(self):
        return False

    def execute(self, entity):
        raise Exception("Never call me")


class DummyPlugin(Plugin):
    """Test class that passes various registration information, and returns a configurable value from execute()."""

    SETTINGS = add_settings(
        # test_input_data allows plugins to pass in data needed in the execute method easily.
        request_retry_count=0,
        test_input_data=(dict, {}),
    )  # Don't retry failed requests when testing
    # leave security property unset
    # SECURITY = None
    VERSION = "1.0"
    MULTI_STREAM_AWARE = True
    FEATURES = [
        Feature("example_string", "Example string feature", type=FeatureType.String),
        Feature("example_int", "Example int feature", type=FeatureType.Integer),
        Feature("example_raw", "Example raw bytes feature", type=FeatureType.Binary),
        Feature("example_date", "Example datetime feature", type=FeatureType.Datetime),
        Feature("example_unspec", "Example feature of unspecified type"),
        Feature("example_path", "Example Filepath feature", type=FeatureType.Filepath),
        Feature("example_uri", "Example URI feature", type=FeatureType.Uri),
    ]

    def execute(self, job: Job):
        pass


class DummyPluginDefaultSharedMem(DummyPlugin):
    def get_shared_memory_name(self):
        raise NotImplementedError(
            "You must define a get_shared_memory_name method that returned the shared memory name"
        )

    def _load_from_shared_memory(self):
        """Unpickle the value stored in shared memory."""
        shared_mem = shared_memory.SharedMemory(name=self.get_shared_memory_name(), create=False)
        return pickle.loads(shared_mem.buf[: shared_mem.size])


def cleanup_shared_memory(shared_memory_name):
    """Delete an old shared memory file if it exists.

    If shutdown happens in a bad way sometimes the shared memory file can get left over.
    """
    try:
        shared_mem_old = shared_memory.SharedMemory(name=shared_memory_name, create=False)
        shared_mem_old.close()
        shared_mem_old.unlink()
    except Exception:
        pass


class DummyPluginFeatureInheritance(DummyPlugin):
    """Test class to ensure that features accumulate between template plugins and their descendants."""

    VERSION = "2.0"  # Much better than 1.0. (Test to ensure it differs from the parent class)
    FEATURES = [
        Feature("descendant feature", "A feature added by the child plugin"),
        Feature("example_unspec", "Child class redefining feature", type=FeatureType.String),
    ]
    # All other class vars should be inherited


class TestPluginSharedMem(TestPlugin):
    def _set_shared_memory(self, shared_mem: shared_memory.SharedMemory, value: Any):
        """Set the shared memory to be equal to the provided value"""
        mem_view = shared_mem.buf
        if not mem_view:
            raise Exception("Memory view is unexpectedly None, can't use shared memory!")
        # Clear memory buffer
        mem_view[: shared_mem.size] = b"\x00" * shared_mem.size
        # Set value
        picked_value = pickle.dumps(value)
        mem_view[: len(picked_value)] = picked_value



"""Dispatcher API Mocking with enough state-tracking to be useful for testing runner."""

import contextlib
import copy
import datetime
import hashlib
import json
import logging
import multiprocessing
import os
import re
import socket
from typing import Any

import urllib3
from azul_bedrock import models_api as azapi
from azul_bedrock import models_network as azm
from azul_bedrock.mock import Editor, MockDispatcher, dp
from fastapi import APIRouter, Request, Response
import uvicorn

router = APIRouter()


class RunnerMockDispatcher(MockDispatcher):
    """Allows for runner to mock dispatcher in forkserver multiprocessing."""

    def run(self, *args, **kwargs):
        """Run the server."""
        app = dp.app
        app.include_router(router)
        config = uvicorn.Config(app, host=self.host, port=self.port)
        self.server = uvicorn.Server(config=config)
        self.config = config
        self.server.run()


class RawResponse(Response):
    media_type = "binary/octet-stream"

    def render(self, content: bytes) -> bytes:
        if not isinstance(content, bytes):
            raise Exception(f"bad response type {type(content)} {content}")
        return content


logger = logging.getLogger(__name__)


REQUEST_RANGE_RE = re.compile(r"bytes=([0-9]*)-([0-9]*)")

# Content for GET /test-mockfile/{foo>
mock_file_data: dict[str, bytes] = {}
posted_files: dict[str, bytes] = {}  # Stores files for later retrieval

vars = {}


def clear():
    global vars, mock_file_data, posted_files
    mock_file_data = {
        "mod256": bytes(range(256)),
    }
    # Stores files for later retrieval
    posted_files = {
        "b4b389c849d799d9331d5937cde7f0dfd297d76083242366cbef53b498cd6051": b"small content",
    }

    # note - not configured/updated for all endpoints
    vars = {
        "last_req_params": {},  # query parameters
        "last_request_body": b"",
        "fetch_count": 0,
        "all_requests": [],
    }


clear()

# ################
# Bottle methods


@router.get("/mock/get_var/{var}")
def mock_get_var(var: str, request: Request, response: Response) -> Any:
    """
    Retrieve a var from the mock server; used to check correct parameters were passed.
    :param var: Name of var
    :return: Value of the var
    """
    response.status_code = 200
    try:
        if var == "all_requests":
            # Format response
            resp = "["
            for val in vars[var]:
                resp += val.decode("utf-8") + ","
            resp = resp[:-1] + "]"
            # Reset requests
            vars[var] = []
            content = resp.encode()
        else:
            content = vars[var]

    except KeyError:
        response.status_code = 400
        return

    if not isinstance(content, bytes):
        content = json.dumps(content).encode()

    return RawResponse(content, status_code=response.status_code, headers=response.headers)


#
# events
#


@router.post("/{path}/api/v2/event")
async def post_event(path: str, request: Request, response: Response) -> azapi.ResponsePostEvent:
    """Called to create new output events on plugin completion"""
    body = await request.body()
    # logger.debug(f"body is {body}")
    vars["last_request_body"] = body
    vars["all_requests"] += [body]

    data = json.loads(body)

    # mock processing all events successfully
    return azapi.ResponsePostEvent(total_ok=len(data), total_failures=0, failures=[])


@router.get("/{variant}/api/v2/event/{ent_type}/active")
async def get_events_switchable(
    variant: str, ent_type: str, request: Request, response: Response
) -> dict | bytes | str:
    """Returns events based on required variant."""
    vars["fetch_count"] += 1
    # save params multi items data structure into dict of list of str
    params = {}
    for k, v in request.query_params.multi_items():
        params.setdefault(k, []).append(v)
    vars["last_req_params"] = params

    respInfo = azapi.GetEventsInfo(filtered=5, fetched=1, ready=True)
    respEvents = azapi.GetEventsBinary(
        events=[
            azm.BinaryEvent(
                model_version=azm.CURRENT_MODEL_VERSION,
                kafka_key="test-dummy",
                dequeued="test-dummy-dequeued",
                action=azm.BinaryAction.Enriched,
                timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                source=azm.Source(
                    name="source",
                    path=[],
                    timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                ),
                author=azm.Author(name="TestServer", category="blah"),
                entity=azm.BinaryEvent.Entity(sha256="1234", datastreams=[], features=[]),
            )
        ],
    )
    if ent_type == "binary":
        if variant == "test_path":
            # returns a binary event message with path
            path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data/binary_event_message.json")
            respEvents.events = json.loads(open(path, "rb").read())
        elif variant == "test_data":
            # returns an entity with data streams
            respEvents.events[0].entity.datastreams = [
                azm.Datastream(
                    identify_version=1,
                    label=azm.DataLabel.CONTENT,
                    size=13,
                    sha256=hashlib.sha256(b"small content").hexdigest(),
                    sha1="1",
                    md5="5",
                    sha512="512",
                    mime="mt",
                    magic="mm",
                    file_format="#TEST/ONLY",
                ),
                azm.Datastream(
                    identify_version=1,
                    label=azm.DataLabel.TEST,
                    size=54,
                    sha256="ed12278c588dcf2932a7db54cf917d8d4da74ce810cf8b7935112174ff2e0df6",
                    # hash of b'This is test data that should be fetched by the runner'
                    # other hashes are not currently used for identification
                    sha1="1",
                    md5="5",
                    sha512="512",
                    mime="mt",
                    magic="mm",
                    file_format="#TEST/ONLY",
                ),
            ]
            respEvents.events[0].source.path = [
                azm.PathNode(
                    author=azm.Author(name="foo"),
                    action=azm.BinaryAction.Extracted,
                    timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                    sha256="1234",
                ),
            ]
        elif variant == "depth_dupe":
            # returns a dummy entity with a source history containing the same author twice
            respEvents.events[0].source.path = [
                azm.PathNode(
                    author=azm.Author(name="foo"),
                    action=azm.BinaryAction.Enriched,
                    timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                    sha256="42",
                ),
                azm.PathNode(
                    author=azm.Author(name="foo"),
                    action=azm.BinaryAction.Sourced,
                    timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                    sha256="1234",
                ),
            ]
            respEvents.events[0].entity.features = [azm.FeatureValue(name="tester", type="integer", value="100")]
        elif variant == "depth_2":
            # returns a dummy entity with a source history two deep from different authors
            respEvents.events[0].source.path = [
                azm.PathNode(
                    author=azm.Author(name="foo"),
                    action=azm.BinaryAction.Enriched,
                    timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                    sha256="42",
                ),
                azm.PathNode(
                    author=azm.Author(name="bar"),
                    action=azm.BinaryAction.Sourced,
                    timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                    sha256="1234",
                ),
            ]
            respEvents.events[0].entity.features = [azm.FeatureValue(name="tester", type="integer", value="100")]
        elif variant == "depth_1":
            # returns a dummy entity with a source history one deep
            respEvents.events[0].source.path = [
                azm.PathNode(
                    author=azm.Author(name="foo"),
                    action=azm.BinaryAction.Enriched,
                    timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                    sha256="1234",
                ),
            ]
            respEvents.events[0].entity.features = [azm.FeatureValue(name="tester", type="integer", value="100")]
        elif variant == "null_sleep":
            # returns a dummy entity with a test feature
            respEvents.events[0].entity.features = [azm.FeatureValue(name="tester", type="integer", value="100")]
        elif variant == "null":
            pass
        elif variant == "invalid":
            respEvents = "This is not valid JSON. This should cause an exception."
        elif variant == "too_many":
            respEvents.events.append(copy.deepcopy(respEvents.events[0]))
        else:
            response.status_code = 404
        content, response.headers["Content-Type"] = urllib3.encode_multipart_formdata(
            {
                "info": (None, respInfo.model_dump_json()),
                "events": (None, respEvents.model_dump_json()),
            }
        )
        return Response(content=content, media_type=response.headers["Content-Type"])
    else:
        response.status_code = 404



from __future__ import annotations

import datetime
import unittest

from pydantic import ValidationError

from azul_runner.models import Feature, FeatureValue

# ##################################
# FeatureValue and related classes


class TestFeatureComparison(unittest.TestCase):
    """Ensure that Feature class behaves as expected - comparison by name only"""

    def test_comparison(self):
        self.assertEqual(Feature("foo", "First description"), Feature("foo", "Other description"))
        self.assertLess(Feature("a feature", ""), Feature("other feature", ""))

    def test_set(self):
        # Same-named features are considered identical so sets should not update with them
        s = {Feature("a", "Feature a"), Feature("b", "Feature b")}
        s.add(Feature("a", "New description"))
        self.assertEqual(s, {Feature("a", "Feature a"), Feature("b", "Feature b")})
        s.update([Feature("a", "Another description")])
        self.assertEqual(s, {Feature("a", "Feature a"), Feature("b", "Feature b")})


class TestFeatureValue(unittest.TestCase):
    """Tests the FeatureValue class, including its label values"""

    def test_comparison(self):
        """Test that the __eq__/__lt__ functions work correctly"""
        self.assertEqual(FeatureValue(1), FeatureValue(1, offset=None, size=None))
        self.assertLess(FeatureValue(5), FeatureValue(10))
        self.assertGreater(FeatureValue("Zoo"), FeatureValue("Alphabet"))
        self.assertLessEqual(FeatureValue(5), FeatureValue(10))
        self.assertLess(FeatureValue(b"bytes"), FeatureValue("a string"))  # 'bytes' < 'str'
        self.assertGreater(
            FeatureValue(56), FeatureValue(datetime.datetime.now(tz=datetime.timezone.utc))
        )  # 'int' > 'datetime'
        # 'None' offset (and size) should be 'less than' any value
        self.assertLess(FeatureValue(0), FeatureValue(0, offset=-9001))
        # Labels should affect comparison
        self.assertNotEqual(FeatureValue(b"foo", label="x"), FeatureValue(b"foo", label="y"))
        self.assertGreater(FeatureValue(1, label="l1", offset=5), FeatureValue(1, label="l1"))
        self.assertLess(FeatureValue(1, label="l1", offset=3, size=2), FeatureValue(1, label="l1", offset=3, size=5))

    def test_sorting(self):
        self.assertEqual(
            sorted(
                [
                    FeatureValue(19),
                    FeatureValue(5),
                    FeatureValue("string"),
                    FeatureValue(b"blob"),
                    FeatureValue("apple"),
                ]
            ),
            [FeatureValue(b"blob"), FeatureValue(5), FeatureValue(19), FeatureValue("apple"), FeatureValue("string")],
        )
        # Test combination of labels and tags
        self.assertEqual(
            sorted(
                [
                    FeatureValue(5, label="two", size=9),
                    FeatureValue(5),
                    FeatureValue(5, label="one"),
                    FeatureValue(5, label="one", offset=55),
                ]
            ),
            [
                FeatureValue(5),
                FeatureValue(5, label="one"),
                FeatureValue(5, label="one", offset=55),
                FeatureValue(5, label="two", size=9),
            ],
        )

    def test_value_validated(self):
        """Tests that FeatureValue.value is forced to be a valid feature data type"""
        for v in (500, "string", b"bytes", datetime.datetime.now(tz=datetime.timezone.utc)):
            # Should raise no error
            _ = FeatureValue(v)
        self.assertRaisesRegex(ValidationError, "Input should be a valid", FeatureValue, None)
        self.assertRaisesRegex(ValidationError, "validation errors", FeatureValue, [1, 2])
        self.assertRaisesRegex(ValidationError, "validation errors", FeatureValue, (3 + 2j))

    def test_hashing(self):
        """Test that FeatureValues work as dict keys / set members"""
        s = {FeatureValue(5), FeatureValue(5, label="l"), FeatureValue("cat"), FeatureValue(b"foo", label="l")}
        self.assertIn(FeatureValue(5), s)
        self.assertIn(FeatureValue(5, label="l"), s)
        self.assertIn(FeatureValue("cat"), s)
        # Value should not match with different tags and labels, unless using match_value_only
        self.assertNotIn(FeatureValue("cat", label="some label"), s)
        self.assertNotIn(FeatureValue(b"foo"), s)
        # Labels etc are distinguished from each other
        self.assertNotEqual(
            {FeatureValue(5), FeatureValue(5, label="f")},
            {FeatureValue(5)},
        )
        # Tags/labels contribute to comparison but not hashing
        # self.assertEqual(
        #     {FeatureValue(5), FeatureValue(5, label='f')},
        #     {FeatureValue(5).match_value_only},
        # )



from __future__ import annotations

import datetime
import hashlib
import time
import unittest
from typing import ClassVar
from unittest import mock

import httpx
import pendulum
from azul_bedrock import models_network as azm

from azul_runner import FeatureType, Job, network_transform, storage
from azul_runner.plugin import Multiplugin
from azul_runner.pusher import Pusher
from tests import plugin_support as sup

from . import mock_dispatcher as md
from . import plugin_support as sup


class TestPusherLive(unittest.TestCase):
    """
    Test cases for base plugin class - cases that talk to the mock server
    """

    mock_server: ClassVar[md.MockDispatcher]
    server: ClassVar[str]  # Endpoint to the mock server, suitable for passing to a plugin's config['server']
    maxDiff = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.mock_server, cls.server = sup.setup_mock_dispatcher()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.mock_server.stop()
        cls.mock_server.kill()

    def get_last_req_body(self) -> dict:
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        return r.json()

    @mock.patch("pendulum.now", lambda *args: pendulum.parse("2023-10-10T10:10:10Z"))
    def test_push_once_sourced(self):
        class DummyPluginWithFeats(sup.DummyPlugin):
            """Plugin that adds features based on input"""

            _IS_USING_PUSHER = True

            def execute(self, job: Job):
                data = job.get_data()
                self.add_feature_values("example_string", "dummy string")
                if data:
                    self.add_feature_values("example_raw", data.read())

        test_source = "source"
        # First attempt at using pusher
        pusher = Pusher(
            DummyPluginWithFeats,
            {
                "events_url": self.server + "/generic",
                "data_url": self.server,
            },
        )

        pusher.push_once_sourced(
            content=b"test-file1",
            security="OFFICIAL",
            source_label=test_source,
            references={"ref1": "val1"},
            filename="random-file-name1",
        )
        result = azm.StatusEvent.model_validate(self.get_last_req_body()[0])
        print(result.entity.results[0].entity.features)
        self.assertCountEqual(
            result.entity.results[0].entity.features,
            [
                azm.FeatureValue(name="example_raw", type=azm.FeatureType.Binary, value="dGVzdC1maWxlMQ=="),
                azm.FeatureValue(name="example_string", type=azm.FeatureType.String, value="dummy string"),
                azm.FeatureValue(name="file_format", type=azm.FeatureType.String, value="#TEST/ONLY"),
                azm.FeatureValue(name="file_extension", type=azm.FeatureType.String, value="tonly"),
                azm.FeatureValue(name="magic", type=azm.FeatureType.String, value="#TESTONLY"),
                azm.FeatureValue(name="mime", type=azm.FeatureType.String, value="#TESTONLY"),
                azm.FeatureValue(name="filename", type=azm.FeatureType.Filepath, value="random-file-name1"),
            ],
        )

        # More data to verify that state isn't going to break anything.
        pusher.push_once_sourced(
            content=b"random-different-data",
            security="OFFICIAL",
            source_label=test_source,
            references={"ref1": "val1"},
            filename="random-diff-data",
        )
        status = self.get_last_req_body()[0]
        result = azm.StatusEvent.model_validate(status)
        print(result.entity.results[0].entity.features)
        self.assertCountEqual(
            result.entity.results[0].entity.features,
            [
                azm.FeatureValue(
                    name="example_raw", type=azm.FeatureType.Binary, value="cmFuZG9tLWRpZmZlcmVudC1kYXRh"
                ),
                azm.FeatureValue(name="example_string", type=azm.FeatureType.String, value="dummy string"),
                azm.FeatureValue(name="file_format", type=azm.FeatureType.String, value="#TEST/ONLY"),
                azm.FeatureValue(name="file_extension", type=azm.FeatureType.String, value="tonly"),
                azm.FeatureValue(name="magic", type=azm.FeatureType.String, value="#TESTONLY"),
                azm.FeatureValue(name="mime", type=azm.FeatureType.String, value="#TESTONLY"),
                azm.FeatureValue(name="filename", type=azm.FeatureType.Filepath, value="random-diff-data"),
            ],
        )

        status["timestamp"] = "2023-10-10T10:10:10+00:00"
        status["entity"]["results"][0]["timestamp"] = "2023-10-10T10:10:10+00:00"
        # Should only be a single event.
        self.assertEqual(len(result.entity.results[0].source.path), 1)
        print(f"actual\n\n{status}\n\nend actual")
        self.assertEqual(
            status,
            {
                "model_version": azm.CURRENT_MODEL_VERSION,
                "kafka_key": "runner-placeholder",
                "timestamp": "2023-10-10T10:10:10+00:00",
                "author": {"category": "plugin", "name": "DummyPluginWithFeats", "version": "1.0"},
                "entity": {
                    "input": {
                        "model_version": azm.CURRENT_MODEL_VERSION,
                        "kafka_key": "DummyPluginWithFeats-placeholder",
                        "timestamp": "2023-10-10T10:10:10+00:00",
                        "author": {
                            "category": "plugin",
                            "name": "DummyPluginWithFeats",
                            "version": "1.0",
                            "security": "OFFICIAL",
                        },
                        "entity": {
                            "sha256": "2f094deac91260d23d3a9a9c0c9e59448342f346fa2927f0ec10af1d72cf55d0",
                            "sha512": "ee543f098404b991962a3072ff0038b1eaf0fbd6d957fdbe6074f9c57d88da9f3546925aedeff4a1917a6cf104aa72e7201069c1202a6e77d0ead590dbb053ea",
                            "sha1": "b8f17c427f04225815f1e2fa1050e3d24c96e550",
                            "md5": "a548bfb3e683426097217764db88e07a",
                            "size": 21,
                            "file_format": "#TEST/ONLY",
                            "file_extension": "tonly",
                            "mime": "#TESTONLY",
                            "magic": "#TESTONLY",
                        },
                        "action": "sourced",
                        "source": {
                            "name": "source",
                            "path": [
                                {
                                    "sha256": "2f094deac91260d23d3a9a9c0c9e59448342f346fa2927f0ec10af1d72cf55d0",
                                    "action": "sourced",
                                    "timestamp": "2023-10-10T10:10:10+00:00",
                                    "author": {
                                        "category": "plugin",
                                        "name": "DummyPluginWithFeats",
                                        "version": "1.0",
                                        "security": "OFFICIAL",
                                    },
                                    "file_format": "#TEST/ONLY",
                                    "size": 21,
                                    "filename": "random-diff-data",
                                }
                            ],
                            "timestamp": "2023-10-10T10:10:10+00:00",
                            "security": "OFFICIAL",
                            "references": {"ref1": "val1"},
                        },
                        "dequeued": "2f094deac91260d23d3a9a9c0c9e59448342f346fa2927f0ec10af1d72cf55d0.DummyPluginWithFeats.1.0.2023-10-10T10:10:10Z",
                    },
                    "status": "completed",
                    "runtime": 0.0,
                    "results": [
                        {
                            "model_version": azm.CURRENT_MODEL_VERSION,
                            "kafka_key": "runner-placeholder",
                            "timestamp": "2023-10-10T10:10:10+00:00",
                            "author": {"category": "plugin", "name": "DummyPluginWithFeats", "version": "1.0"},
                            "entity": {
                                "sha256": "2f094deac91260d23d3a9a9c0c9e59448342f346fa2927f0ec10af1d72cf55d0",
                                "sha512": "ee543f098404b991962a3072ff0038b1eaf0fbd6d957fdbe6074f9c57d88da9f3546925aedeff4a1917a6cf104aa72e7201069c1202a6e77d0ead590dbb053ea",
                                "sha1": "b8f17c427f04225815f1e2fa1050e3d24c96e550",
                                "md5": "a548bfb3e683426097217764db88e07a",
                                "size": 21,
                                "file_format": "#TEST/ONLY",
                                "file_extension": "tonly",
                                "mime": "#TESTONLY",
                                "magic": "#TESTONLY",
                                "features": [
                                    {"name": "example_raw", "type": "binary", "value": "cmFuZG9tLWRpZmZlcmVudC1kYXRh"},
                                    {"name": "example_string", "type": "string", "value": "dummy string"},
                                    {"name": "file_format", "type": "string", "value": "#TEST/ONLY"},
                                    {"name": "file_extension", "type": "string", "value": "tonly"},
                                    {"name": "magic", "type": "string", "value": "#TESTONLY"},
                                    {"name": "mime", "type": "string", "value": "#TESTONLY"},
                                    {"name": "filename", "type": "filepath", "value": "random-diff-data"},
                                ],
                                "datastreams": [
                                    {
                                        "sha256": "2f094deac91260d23d3a9a9c0c9e59448342f346fa2927f0ec10af1d72cf55d0",
                                        "sha512": "ee543f098404b991962a3072ff0038b1eaf0fbd6d957fdbe6074f9c57d88da9f3546925aedeff4a1917a6cf104aa72e7201069c1202a6e77d0ead590dbb053ea",
                                        "sha1": "b8f17c427f04225815f1e2fa1050e3d24c96e550",
                                        "md5": "a548bfb3e683426097217764db88e07a",
                                        "size": 21,
                                        "file_format": "#TEST/ONLY",
                                        "file_extension": "tonly",
                                        "mime": "#TESTONLY",
                                        "magic": "#TESTONLY",
                                        "identify_version": 1,
                                        "label": "content",
                                    }
                                ],
                            },
                            "action": "sourced",
                            "source": {
                                "name": "source",
                                "path": [
                                    {
                                        "sha256": "2f094deac91260d23d3a9a9c0c9e59448342f346fa2927f0ec10af1d72cf55d0",
                                        "action": "sourced",
                                        "timestamp": "2023-10-10T10:10:10+00:00",
                                        "author": {
                                            "category": "plugin",
                                            "name": "DummyPluginWithFeats",
                                            "version": "1.0",
                                            "security": "OFFICIAL",
                                        },
                                        "file_format": "#TEST/ONLY",
                                        "size": 21,
                                        "filename": "random-diff-data",
                                    }
                                ],
                                "timestamp": "2023-10-10T10:10:10+00:00",
                                "security": "OFFICIAL",
                                "references": {"ref1": "val1"},
                            },
                        }
                    ],
                },
            },
        )

    @mock.patch("pendulum.now", lambda *args: pendulum.parse("2023-10-10T10:10:10Z"))
    def test_push_once_mapped(self):
        class DummyPluginWithFeats(sup.DummyPlugin):
            """Plugin that adds features based on input"""

            _IS_USING_PUSHER = True

            def execute(self, job: Job):
                data = job.get_data()
                self.add_feature_values("example_string", "dummy string")
                if data:
                    self.add_feature_values("example_raw", data.read())

        test_source = "source"
        # First attempt at using pusher
        pusher = Pusher(
            DummyPluginWithFeats,
            {
                "events_url": self.server + "/generic",
                "data_url": self.server,
            },
        )

        source_info = pusher.generate_base_mapped_source(
            source_label=test_source, references={"ref1": "val1"}, security="OFFICIAL"
        )
        pusher.push_once_mapped(
            source_file_info=azm.FileInfo(
                sha256="dummy-sha256",
                sha512="dummy-sha512",
                size=1,
                file_format="text/plain",
            ),
            source_info=source_info,
            filename="random-file-name1",
            security="OFFICIAL",
        )
        result = azm.StatusEvent.model_validate(self.get_last_req_body()[0])
        print(result.entity.results[0].entity.features)
        self.assertCountEqual(
            result.entity.results[0].entity.features,
            [
                azm.FeatureValue(name="example_string", type=azm.FeatureType.String, value="dummy string"),
                azm.FeatureValue(name="filename", type=azm.FeatureType.Filepath, value="random-file-name1"),
                azm.FeatureValue(name="file_format", type=azm.FeatureType.String, value="text/plain"),
            ],
        )

        # Again generating a mapped event instead of a sourced event by providing no data
        # Results in less features in the sourced event and one less from the plugin itself.
        source_info = pusher.generate_base_mapped_source(
            source_label=test_source, references={"ref1": "val1"}, security="OFFICIAL"
        )
        source_info.path.append(
            azm.PathNode(
                sha256="random-other-sha256",
                action=azm.BinaryAction.Sourced,
                timestamp=pendulum.now(),
                author=azm.Author(
                    name="user1",
                    category="User",
                    security="OFFICIAL",
                ),
                file_format="text/plain",
            )
        )
        pusher.push_once_mapped(
            source_file_info=azm.FileInfo(
                sha256="dummy-sha256",
                sha512="dummy-sha512",
                size=1,
                file_format="text/plain",
            ),
            source_info=source_info,
            security="OFFICIAL",
            filename="random-file-name1",
            relationship={"Asssemblyline_reingest": "Reingested from Assemblyline"},
        )
        status = self.get_last_req_body()[0]
        result = azm.StatusEvent.model_validate(status)
        print(result.entity.results[0].entity.features)
        self.assertCountEqual(
            result.entity.results[0].entity.features,
            [
                azm.FeatureValue(name="example_string", type=azm.FeatureType.String, value="dummy string"),
                azm.FeatureValue(name="filename", type=azm.FeatureType.Filepath, value="random-file-name1"),
                azm.FeatureValue(name="file_format", type=azm.FeatureType.String, value="text/plain"),
            ],
        )

        status["timestamp"] = "2023-10-10T10:10:10+00:00"
        status["entity"]["results"][0]["timestamp"] = "2023-10-10T10:10:10+00:00"
        # Should be attached to provided source info so should have depth of two.
        self.assertEqual(len(result.entity.results[0].source.path), 2)

        print(f"actual\n\n{status}\n\nend actual")
        self.assertEqual(
            status,
            {
                "model_version": azm.CURRENT_MODEL_VERSION,
                "kafka_key": "runner-placeholder",
                "timestamp": "2023-10-10T10:10:10+00:00",
                "author": {"category": "plugin", "name": "DummyPluginWithFeats", "version": "1.0"},
                "entity": {
                    "input": {
                        "model_version": azm.CURRENT_MODEL_VERSION,
                        "kafka_key": "DummyPluginWithFeats-placeholder",
                        "timestamp": "2023-10-10T10:10:10+00:00",
                        "author": {
                            "category": "plugin",
                            "name": "DummyPluginWithFeats",
                            "version": "1.0",
                            "security": "OFFICIAL",
                        },
                        "entity": {
                            "sha256": "dummy-sha256",
                            "sha512": "dummy-sha512",
                            "size": 1,
                            "file_format": "text/plain",
                        },
                        "action": "mapped",
                        "source": {
                            "security": "OFFICIAL",
                            "name": "source",
                            "timestamp": "2023-10-10T10:10:10+00:00",
                            "references": {"ref1": "val1"},
                            "path": [
                                {
                                    "sha256": "random-other-sha256",
                                    "action": "sourced",
                                    "timestamp": "2023-10-10T10:10:10+00:00",
                                    "author": {"category": "User", "name": "user1", "security": "OFFICIAL"},
                                    "file_format": "text/plain",
                                },
                                {
                                    "sha256": "dummy-sha256",
                                    "action": "mapped",
                                    "timestamp": "2023-10-10T10:10:10+00:00",
                                    "author": {
                                        "category": "plugin",
                                        "name": "DummyPluginWithFeats",
                                        "version": "1.0",
                                        "security": "OFFICIAL",
                                    },
                                    "relationship": {"Asssemblyline_reingest": "Reingested from Assemblyline"},
                                    "file_format": "text/plain",
                                    "size": 1,
                                    "filename": "random-file-name1",
                                },
                            ],
                        },
                        "dequeued": "dummy-sha256.DummyPluginWithFeats.1.0.2023-10-10T10:10:10Z",
                    },
                    "status": "completed",
                    "runtime": 0.0,
                    "results": [
                        {
                            "model_version": azm.CURRENT_MODEL_VERSION,
                            "kafka_key": "runner-placeholder",
                            "timestamp": "2023-10-10T10:10:10+00:00",
                            "author": {"category": "plugin", "name": "DummyPluginWithFeats", "version": "1.0"},
                            "entity": {
                                "sha256": "dummy-sha256",
                                "sha512": "dummy-sha512",
                                "size": 1,
                                "file_format": "text/plain",
                                "features": [
                                    {"name": "example_string", "type": "string", "value": "dummy string"},
                                    {"name": "file_format", "type": "string", "value": "text/plain"},
                                    {"name": "filename", "type": "filepath", "value": "random-file-name1"},
                                ],
                            },
                            "action": "mapped",
                            "source": {
                                "security": "OFFICIAL",
                                "name": "source",
                                "timestamp": "2023-10-10T10:10:10+00:00",
                                "references": {"ref1": "val1"},
                                "path": [
                                    {
                                        "sha256": "random-other-sha256",
                                        "action": "sourced",
                                        "timestamp": "2023-10-10T10:10:10+00:00",
                                        "author": {"category": "User", "name": "user1", "security": "OFFICIAL"},
                                        "file_format": "text/plain",
                                    },
                                    {
                                        "sha256": "dummy-sha256",
                                        "action": "mapped",
                                        "timestamp": "2023-10-10T10:10:10+00:00",
                                        "author": {
                                            "category": "plugin",
                                            "name": "DummyPluginWithFeats",
                                            "version": "1.0",
                                            "security": "OFFICIAL",
                                        },
                                        "relationship": {"Asssemblyline_reingest": "Reingested from Assemblyline"},
                                        "file_format": "text/plain",
                                        "size": 1,
                                        "filename": "random-file-name1",
                                    },
                                ],
                            },
                        }
                    ],
                },
            },
        )

    @mock.patch("pendulum.now", lambda *args: pendulum.parse("2023-10-10T10:10:10Z"))
    def test_push_once_alt_stream(self):
        class DummyPluginWithAlt(sup.DummyPlugin):
            """Plugin that adds features based on input"""

            _IS_USING_PUSHER = True

            def execute(self, job: Job):
                data = job.get_data()
                data_alt = job.get_data(azm.DataLabel.TEST)
                self.add_feature_values("example_raw", [data.read(), data_alt.read()])

        test_source = "source"
        pusher = Pusher(
            DummyPluginWithAlt,
            {
                "events_url": self.server + "/generic",
                "data_url": self.server,
            },
        )

        binary_data = b"test-file1"
        alt_content = b"alt-stream-data-is-here-and-available."
        with storage.StorageProxyFile(
            source="alt-source",
            label=azm.DataLabel.TEST,
            hash=hashlib.sha256(alt_content).hexdigest(),
            init_data=alt_content,
            file_info=azm.Datastream(label=azm.DataLabel.TEST, file_format="text/plain"),
            allow_unbounded_read=True,
        ) as meta_spf:
            with storage.StorageProxyFile(
                source=test_source,
                label=azm.DataLabel.CONTENT,
                hash=hashlib.sha256(binary_data).hexdigest(),
                init_data=binary_data,
                file_info=azm.Datastream(label=azm.DataLabel.CONTENT, file_format="text/plain"),
                allow_unbounded_read=True,
            ) as spf:
                pusher.push_once_sourced(
                    security="OFFICIAL",
                    source_label=test_source,
                    references={"ref1": "val1"},
                    filename="random-file-name1",
                    content=binary_data,
                    local=[spf, meta_spf],
                )
        result = azm.StatusEvent.model_validate(self.get_last_req_body()[0])
        print(result.entity.results[0].entity.features)
        self.assertCountEqual(
            result.entity.results[0].entity.features,
            [
                azm.FeatureValue(
                    name="example_raw",
                    type=azm.FeatureType.Binary,
                    value="YWx0LXN0cmVhbS1kYXRhLWlzLWhlcmUtYW5kLWF2YWlsYWJsZS4=",
                ),
                azm.FeatureValue(name="example_raw", type=azm.FeatureType.Binary, value="dGVzdC1maWxlMQ=="),
                azm.FeatureValue(name="file_format", type=azm.FeatureType.String, value="#TEST/ONLY"),
                azm.FeatureValue(name="file_extension", type=azm.FeatureType.String, value="tonly"),
                azm.FeatureValue(name="magic", type=azm.FeatureType.String, value="#TESTONLY"),
                azm.FeatureValue(name="mime", type=azm.FeatureType.String, value="#TESTONLY"),
                azm.FeatureValue(name="filename", type=azm.FeatureType.Filepath, value="random-file-name1"),
            ],
        )



"""Subclass of Plugin designed for processing binary files."""

from azul_bedrock import models_network

from . import plugin, settings


class BinaryPlugin(plugin.Plugin):
    """Base class for plugins that want to run on new sightings of a binary.

    See readme.md for usage information.
    """

    SETTINGS = settings.add_settings(
        # plugin code generally assumes it can access data
        assume_streams_available=True,
        # event must be a new sighting of a label='content' binary
        filter_allow_event_types=[
            models_network.BinaryAction.Sourced,
            models_network.BinaryAction.Extracted,
        ],
    )



from __future__ import annotations

import hashlib
import io
import json
import logging
import os
import sys
import tempfile
import unittest
from multiprocessing import managers, shared_memory
from typing import Any, Type
from unittest import mock

from azul_bedrock import models_network as azm

from azul_runner import (
    Feature,
    FeatureType,
    FeatureValue,
    Job,
    Plugin,
    add_settings,
    cmdline_run,
    main,
    monitor,
)

# Name used to pass the shareable list between running processes.
COMMON_NAME_FOR_SHARED_MEM = "TestPluginBinaryInput"


class MainBaseTestPlugin(Plugin):
    SETTINGS = add_settings()


class TestCmdlineRun(unittest.TestCase):
    maxDiff = None

    # ####################### Utility method
    @staticmethod
    def _stdout_result(plugin: Type[Plugin]) -> str:
        saved_stdout = sys.stdout
        stdbuf = sys.stdout = io.StringIO()
        cmdline_run(plugin=plugin)
        sys.stdout = saved_stdout
        return stdbuf.getvalue()

    @staticmethod
    def _stdout_stderr_result(plugin: Type[Plugin]) -> tuple[str, str, str]:
        """Capture stdout, stderr and logs."""
        logbuf = io.StringIO()

        logger = logging.getLogger("azul.plugin")
        logger.setLevel(logging.DEBUG)
        handler = logging.StreamHandler(logbuf)
        handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        saved_stdout = sys.stdout
        saved_stderr = sys.stderr
        stdbuf = sys.stdout = io.StringIO()
        errbuf = sys.stderr = io.StringIO()
        try:
            cmdline_run(plugin=plugin)
        except SystemExit:
            pass
        sys.stdout = saved_stdout
        sys.stderr = saved_stderr
        return stdbuf.getvalue(), errbuf.getvalue(), logbuf.getvalue()

    # ######################## Test methods

    class TestPluginSimplePlugin(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [Feature("value", "A return value", type=FeatureType.Integer)]

        def execute(self, job: Job) -> dict:
            self.add_feature_values("value", 97)

    def test_simple_plugin(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"blah")
            f.seek(0)
            sha256 = hashlib.sha256(f.read()).hexdigest()
            f.seek(0)
            sys.argv = ["test_main.py", f.name]
            res = self._stdout_result(self.TestPluginSimplePlugin)
        self.assertEqual(
            res,
            f"""----- TestPluginSimplePlugin results -----
COMPLETED

events (1)

event for {sha256}:None
  {{}}
  output features:
    value: 97

Feature key:
  value:  A return value

""",
        )

    class TestPluginSimplePluginEmptyFeatureVals(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [Feature("value", "A return value", type=FeatureType.Integer)]

        def execute(self, job: Job) -> dict:
            self.add_feature_values("value", [])
            self.add_feature_values("value", set())

    def test_simple_plugin_empty_feature_value(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"blah")
            f.seek(0)
            sha256 = hashlib.sha256(f.read()).hexdigest()
            f.seek(0)
            sys.argv = ["test_main.py", f.name]
            res = self._stdout_result(self.TestPluginSimplePluginEmptyFeatureVals)
        self.assertEqual(res, "----- TestPluginSimplePluginEmptyFeatureVals results -----\nCOMPLETED_EMPTY\n\n")

    class TestPluginSimplePluginOffsetFeatureValues(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [Feature("value", "A return value", type=FeatureType.Integer)]

        def execute(self, job: Job) -> dict:
            self.add_feature_values("value", FeatureValue(value=97, label="label1", offset=10, size=1))
            self.add_feature_values("value", FeatureValue(value=97, offset=20))
            self.add_feature_values("value", FeatureValue(value=97, label="labelOnly"))
            self.add_feature_values("value", FeatureValue(value=97, size=99))

    def test_simple_plugin_offset_feature_values(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"blah")
            f.seek(0)
            sha256 = hashlib.sha256(f.read()).hexdigest()
            f.seek(0)
            sys.argv = ["test_main.py", f.name]
            res = self._stdout_result(self.TestPluginSimplePluginOffsetFeatureValues)
        self.assertEqual(
            res,
            f"""----- TestPluginSimplePluginOffsetFeatureValues results -----
COMPLETED

events (1)

event for {sha256}:None
  {{}}
  output features:
    value: 97
           97 @ 0x14 (offset)
           label1 - 97 @ 0xa (offset)
           labelOnly - 97

Feature key:
  value:  A return value

""",
        )

    class TestPluginSimplePluginLogDebug(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [Feature("value", "A return value", type=FeatureType.Integer)]

        def execute(self, job: Job) -> dict:
            self.logger.error("a error thing happened")
            self.logger.warning("a warning thing happened")
            self.logger.info("a info thing happened")
            self.logger.debug("a debug thing happened")
            self.add_feature_values("value", 97)

            mylog = logging.getLogger("external.flatlib")
            mylog.error("this is just a simple error")
            mylog.warning("this is just a simple warning")
            mylog.info("this is just a simple info")
            mylog.debug("this is just a simple debug")

    def test_simple_plugin_log_debug(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"blah")
            f.seek(0)
            sha256 = hashlib.sha256(f.read()).hexdigest()
            f.seek(0)
            sys.argv = ["test_main.py", f.name, "--verbose", "--verbose"]
            res, err, log = self._stdout_stderr_result(self.TestPluginSimplePluginLogDebug)
        self.assertEqual(
            res,
            f"""----- TestPluginSimplePluginLogDebug results -----
COMPLETED

events (1)

event for {sha256}:None
  {{}}
  output features:
    value: 97

Feature key:
  value:  A return value

""",
        )
        # Show logs incase of errors
        print(f"--- error ---\n{err}\n--- End error ---")
        print(f"+++ log +++\n{log}\n+++ End log +++")
        # runner logs
        self.assertRegex(err, r"azul.plugin.testplugin.*a error thing happened")
        self.assertRegex(err, r"azul.plugin.testplugin.*a warning thing happened")
        self.assertRegex(err, r"azul.plugin.testplugin.*a info thing happened")
        self.assertRegex(err, r"azul.plugin.testplugin.*a debug thing happened")
        # external tool logs
        self.assertRegex(err, r"external.flatlib.*this is just a simple error")
        self.assertRegex(err, r"external.flatlib.*this is just a simple warning")
        self.assertRegex(err, r"external.flatlib.*this is just a simple info")
        self.assertRegex(err, r"external.flatlib.*this is just a simple debug")

    class TestPluginSimplePluginLogInfo(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [Feature("value", "A return value", type=FeatureType.Integer)]

        def execute(self, job: Job) -> dict:
            self.logger.error("a error thing happened")
            self.logger.warning("a warning thing happened")
            self.logger.info("a info thing happened")
            self.logger.debug("a debug thing happened")
            self.add_feature_values("value", 97)

            mylog = logging.getLogger("external.flatlib")
            mylog.error("this is just a simple error")
            mylog.warning("this is just a simple warning")
            mylog.info("this is just a simple info")
            mylog.debug("this is just a simple debug")

    def test_simple_plugin_log_info(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"blah")
            f.seek(0)
            sha256 = hashlib.sha256(f.read()).hexdigest()
            f.seek(0)
            sys.argv = ["test_main.py", f.name, "--verbose"]
            res, err, log = self._stdout_stderr_result(self.TestPluginSimplePluginLogInfo)
        # Show logs incase of errors
        print(f"--- error ---\n{err}\n--- End error ---")
        print(f"+++ log +++\n{log}\n+++ End log +++")
        # runner logs
        self.assertRegex(err, r"azul.plugin.testplugin.*a error thing happened")
        self.assertRegex(err, r"azul.plugin.testplugin.*a warning thing happened")
        self.assertRegex(err, r"azul.plugin.testplugin.*a info thing happened")
        self.assertNotRegex(err, r"azul.plugin.testplugin.*a debug thing happened")
        # external tool logs
        self.assertRegex(err, r"external.flatlib.*this is just a simple error")
        self.assertRegex(err, r"external.flatlib.*this is just a simple warning")
        self.assertRegex(err, r"external.flatlib.*this is just a simple info")
        self.assertNotRegex(err, r"external.flatlib.*this is just a simple debug")

    class TestPluginSimplePluginLogWarn(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [Feature("value", "A return value", type=FeatureType.Integer)]

        def execute(self, job: Job) -> dict:
            self.logger.error("a error thing happened")
            self.logger.warning("a warning thing happened")
            self.logger.info("a info thing happened")
            self.logger.debug("a debug thing happened")
            self.add_feature_values("value", 97)

            mylog = logging.getLogger("external.flatlib")
            mylog.error("this is just a simple error")
            mylog.warning("this is just a simple warning")
            mylog.info("this is just a simple info")
            mylog.debug("this is just a simple debug")

    def test_simple_plugin_log_warn(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"blah")
            f.seek(0)
            sha256 = hashlib.sha256(f.read()).hexdigest()
            f.seek(0)
            sys.argv = ["test_main.py", f.name]
            res, err, log = self._stdout_stderr_result(self.TestPluginSimplePluginLogWarn)
        # Show logs incase of errors
        print(f"--- error ---\n{err}\n--- End error ---")
        print(f"+++ log +++\n{log}\n+++ End log +++")
        # runner logs
        self.assertRegex(err, r"azul.plugin.testplugin.*a error thing happened")
        self.assertRegex(err, r"azul.plugin.testplugin.*a warning thing happened")
        self.assertNotRegex(err, r"azul.plugin.testplugin.*a info thing happened")
        self.assertNotRegex(err, r"azul.plugin.testplugin.*a debug thing happened")
        # external tool logs
        self.assertRegex(err, r"external.flatlib.*this is just a simple error")
        self.assertRegex(err, r"external.flatlib.*this is just a simple warning")
        self.assertNotRegex(err, r"external.flatlib.*this is just a simple info")
        self.assertNotRegex(err, r"external.flatlib.*this is just a simple debug")

    class TestPluginSimpleLogErr(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [Feature("value", "A return value", type=FeatureType.Integer)]

        def execute(self, job: Job) -> dict:
            self.logger.error("a error thing happened")
            self.logger.warning("a warning thing happened")
            self.logger.info("a info thing happened")
            self.logger.debug("a debug thing happened")
            self.add_feature_values("value", 97)

            mylog = logging.getLogger("external.flatlib")
            mylog.error("this is just a simple error")
            mylog.warning("this is just a simple warning")
            mylog.info("this is just a simple info")
            mylog.debug("this is just a simple debug")

    def test_simple_plugin_log_err(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"blah")
            f.seek(0)
            sha256 = hashlib.sha256(f.read()).hexdigest()
            f.seek(0)
            sys.argv = ["test_main.py", f.name, "--quiet"]
            res, err, log = self._stdout_stderr_result(self.TestPluginSimpleLogErr)
        # Show logs incase of errors
        print(f"--- error ---\n{err}\n--- End error ---")
        print(f"+++ log +++\n{log}\n+++ End log +++")
        # runner logs
        self.assertRegex(err, r"azul.plugin.testplugin.*a error thing happened")
        self.assertNotRegex(err, r"azul.plugin.testplugin.*a warning thing happened")
        self.assertNotRegex(err, r"azul.plugin.testplugin.*a info thing happened")
        self.assertNotRegex(err, r"azul.plugin.testplugin.*a debug thing happened")
        # external tool logs
        self.assertRegex(err, r"external.flatlib.*this is just a simple error")
        self.assertNotRegex(err, r"external.flatlib.*this is just a simple warning")
        self.assertNotRegex(err, r"external.flatlib.*this is just a simple info")
        self.assertNotRegex(err, r"external.flatlib.*this is just a simple debug")

    class TestPluginBinaryInput(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [Feature("job_event_id", "Getting Job details back", type=FeatureType.Integer)]

        def execute(self, job: Job) -> dict:
            shared_list = shared_memory.ShareableList(None, name=COMMON_NAME_FOR_SHARED_MEM)
            shared_list[0] = job.event.entity.sha256
            shared_list[1] = job.event.entity.sha512
            shared_list[2] = job.event.entity.sha1
            shared_list[3] = job.event.entity.md5
            shared_list[4] = job.event.entity.size
            shared_list[5] = job.event.entity.file_format

    def test_binary_input(self):
        shared_list = shared_memory.ShareableList([500 * "a" for _ in range(7)], name=COMMON_NAME_FOR_SHARED_MEM)
        try:
            with tempfile.NamedTemporaryFile() as f:
                f.write(b"blah")
                f.seek(0)
                hashlib.sha256(f.read()).hexdigest()
                f.seek(0)
                sys.argv = ["test_main.py", f.name]
                result_data = self._stdout_result(self.TestPluginBinaryInput)
                # Logs errors if something is wrong with the shared list.
                print(result_data)
                result_dict = {
                    "job.event.entity.sha256": shared_list[0],
                    "job.event.entity.sha512": shared_list[1],
                    "job.event.entity.sha1": shared_list[2],
                    "job.event.entity.md5": shared_list[3],
                    "job.event.entity.size": shared_list[4],
                    "job.event.entity.file_format": shared_list[5],
                }
                print(result_dict)
                self.assertEqual(
                    result_dict["job.event.entity.sha256"],
                    "8b7df143d91c716ecfa5fc1730022f6b421b05cedee8fd52b1fc65a96030ad52",
                )
                self.assertEqual(
                    result_dict["job.event.entity.sha512"],
                    "39ca2b1f97c7d1d223dcb2b22cbe20c36f920aeefd201d0bf68ffc08db6d9ac608a0a202fb536d944c9d1f50cf9bd61b5bc84217212f0727a8db8a01c2fa54b7",
                )
                self.assertEqual(result_dict["job.event.entity.sha1"], "5bf1fd927dfb8679496a2e6cf00cbe50c1c87145")
                self.assertEqual(result_dict["job.event.entity.md5"], "6f1ed002ab5595859014ebf0951522d9")
                self.assertEqual(result_dict["job.event.entity.size"], 4)
                self.assertEqual(result_dict["job.event.entity.file_format"], "text/plain")
        finally:
            # Close shared memory
            shared_list.shm.close()
            shared_list.shm.unlink()

    class TestPluginSimplePluginOnlyStream(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [Feature("value", "A return value", FeatureType.String)]

        def execute(self, job: Job) -> dict:
            data = job.get_data(azm.DataLabel.TEST).read().decode("utf-8", errors="ignore")
            self.add_feature_values("value", data)

    def test_simple_plugin_only_stream(self):
        sha256 = "test_stream_only"
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"blah" * 4)
            f.seek(0)
            sys.argv = ["test_main.py", "--entity-id", sha256, "--stream", azm.DataLabel.TEST, f.name]
            res = self._stdout_result(self.TestPluginSimplePluginOnlyStream)
        self.assertEqual(
            res,
            f"""----- TestPluginSimplePluginOnlyStream results -----
COMPLETED

events (1)

event for {sha256}:None
  {{}}
  output features:
    value: blahblahblahblah

Feature key:
  value:  A return value

""",
        )

    class TestPluginMultipleStreams(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [Feature("value", "A return value", type=FeatureType.Integer)]

        def execute(self, job: Job) -> dict:
            text = job.get_data(azm.DataLabel.TEST).file_info.sha256
            event = self.get_data_event(text)
            event.add_feature_values("value", FeatureValue(95))
            self.add_feature_values("value", 97)

    def test_multiple_streams(self):
        # execute again with two input data streams - expect a result for each stream
        saved_argv = sys.argv
        tf1 = tempfile.NamedTemporaryFile()
        tf1.write(b"Not a Content stream")
        tf1.seek(0)
        sha256 = hashlib.sha256(tf1.read()).hexdigest()
        tf1.seek(0)
        tf1.flush()
        tf2 = tempfile.NamedTemporaryFile()
        tf2.write(b"Text stream")
        tf2.seek(0)
        tf2.flush()
        sys.argv = ["test_main.py", "-s", azm.DataLabel.TEXT, tf1.name, "-s", azm.DataLabel.TEST, tf2.name]
        res = self._stdout_result(self.TestPluginMultipleStreams)
        print(res)
        self.assertEqual(
            res,
            """----- TestPluginMultipleStreams results -----
COMPLETED

events (2)

event for cmdline_entity:None
  {}
  output features:
    value: 97

event for data 7b238c15c251a28c21976f22d47d9a7f31c3d8cff922f53c4129eee140bdc673
  {}
  output features:
    value: 95

Feature key:
  value:  A return value

""",
        )
        tf1.close()
        tf2.close()

    class TestPluginChildOutput(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [Feature("child_feat", "A feature of the child entity", FeatureType.String)]

        def execute(self, job: Job) -> dict:
            c = self._add_child(
                "child_ID",
                {"relationship": "extracted"},
            )
            c.add_data("content", {}, b"Child content")
            c.add_feature_values("child_feat", "child feature value")

            c = self._add_child(
                "child2_ID",
                {"relationship": "extracted"},
            )
            c.add_data("content", {}, b"Child content")
            c.add_feature_values("child_feat", "child feature value")
            c = self._add_child(
                "child2_ID",
                {"relationship": "extracted"},
            )
            c.add_data("content", {}, b"Child content")
            c.add_feature_values("child_feat", "child feature value")
            c = self._add_child(
                "child3_ID",
                {"relationship": "extracted"},
            )
            c.add_data("content", {}, b"Child content")
            c.add_feature_values("child_feat", "child feature value")
            c = self._add_child(
                "child3_ID",
                {"relationship": "extracted", "additional": "extracted twice"},
            )
            c.add_data("content", {}, b"Child content")
            c.add_feature_values("child_feat", "a different feature value")
            return

    def test_child_output(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"blah")
            f.seek(0)
            sha256 = hashlib.sha256(f.read()).hexdigest()
            f.seek(0)
            sys.argv = ["test_main.py", f.name]
            res = self._stdout_result(self.TestPluginChildOutput)
            print(res)
        self.assertEqual(
            res,
            f"""----- TestPluginChildOutput results -----
COMPLETED

events (3)

event for child_ID:None
  {{'relationship': 'extracted'}}
  child of {sha256}
  output data streams (1):
    13 bytes - EventData(hash='20ec1767a3be628df0b67c6bd9016ad50d106da0775a3fe65e49ca54121747ea', label='content')
  output features:
    child_feat: child feature value

event for child2_ID:None
  {{'relationship': 'extracted'}}
  child of {sha256}
  output data streams (1):
    13 bytes - EventData(hash='20ec1767a3be628df0b67c6bd9016ad50d106da0775a3fe65e49ca54121747ea', label='content')
  output features:
    child_feat: child feature value

event for child3_ID:None
  {{'relationship': 'extracted'}}
  child of {sha256}
  output data streams (1):
    13 bytes - EventData(hash='20ec1767a3be628df0b67c6bd9016ad50d106da0775a3fe65e49ca54121747ea', label='content')
  output features:
    child_feat: a different feature value
                child feature value

Feature key:
  child_feat:  A feature of the child entity

""",
        )

    class TestPluginWithConfig(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = []

        def execute(self, job: Job) -> dict:
            return

    def test_with_config(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"blah")
            f.seek(0)
            sys.argv = ["test_main.py", f.name, "-c", "alpha", "beta"]
            res = self._stdout_result(self.TestPluginWithConfig)
        print(res)
        self.assertEqual(res, "----- TestPluginWithConfig results -----\nCOMPLETED_EMPTY\n\n")

    class TestPluginGrandchildOutput(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [
            Feature("child_feat", "A feature of the child entity", FeatureType.String),
            Feature("gc_feat", "A feature of the grandchild", FeatureType.String),
        ]

        def execute(self, job: Job) -> dict:
            c = self._add_child(
                "child_ID",
                {"relationship": "extracted"},
            )
            c.add_data("content", {}, b"Child content")
            c.add_feature_values("child_feat", "child feature value")
            gc = c._add_child("grandchild_ID", {"relationship": "decoded"})
            gc.add_data("content", {"language": "ocelot"}, b"Grandchild content")
            gc.add_feature_values("gc_feat", ["value1", "value2"])
            return

    def test_grandchild_output(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"blah")
            f.seek(0)
            sha256 = hashlib.sha256(f.read()).hexdigest()
            f.seek(0)
            sys.argv = ["test_main.py", f.name]
            res = self._stdout_result(self.TestPluginGrandchildOutput)
        print(res)
        self.assertEqual(
            res,
            f"""----- TestPluginGrandchildOutput results -----
COMPLETED

events (2)

event for child_ID:None
  {{'relationship': 'extracted'}}
  child of {sha256}
  output data streams (1):
    13 bytes - EventData(hash='20ec1767a3be628df0b67c6bd9016ad50d106da0775a3fe65e49ca54121747ea', label='content')
  output features:
    child_feat: child feature value

event for grandchild_ID:None
  {{'relationship': 'decoded'}}
  child of child_ID
  output data streams (1):
    18 bytes - EventData(hash='1459c1f57adc3cc9ec8da1dddcce220eac175f048f212a1392325ad112f39422', label='content', language='ocelot')
  output features:
    gc_feat: value1
             value2

Feature key:
  child_feat:  A feature of the child entity
  gc_feat:  A feature of the grandchild

""",
        )

    class TestPluginStreamOutput(MainBaseTestPlugin):
        VERSION = "none"

        def execute(self, job: Job) -> dict:
            self.add_data(
                "text",
                {"language": "python"},
                b"Text output content",
            )
            return

    def test_stream_output(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"blah")
            f.seek(0)
            sha256 = hashlib.sha256(f.read()).hexdigest()
            f.seek(0)
            sys.argv = ["test_main.py", f.name]
            res = self._stdout_result(self.TestPluginStreamOutput)
        print(res)
        self.assertEqual(
            res,
            f"""----- TestPluginStreamOutput results -----
COMPLETED

events (1)

event for {sha256}:None
  {{}}
  output data streams (1):
    19 bytes - EventData(hash='1c2568562238a338919f87598737756a67e22f35d5031a45796c9d312b5beb54', label='text', language='python')

""",
        )

    class TestPluginJsonOutput(MainBaseTestPlugin):
        VERSION = "1.0"
        FEATURES = [Feature("value", "A return value", type=FeatureType.Integer)]

        def execute(self, job: Job) -> dict:
            self.add_data(
                "text",
                {"language": "python"},
                b"Text output content",
            )
            self.add_feature_values("value", 12345)

    def test_JSON_output(self):

        with tempfile.NamedTemporaryFile() as f:
            f.write(b"blah")
            f.seek(0)
            sha256 = hashlib.sha256(f.read()).hexdigest()
            f.seek(0)
            sys.argv = ["test_main.py", f.name, "--output-json"]
            res = self._stdout_result(self.TestPluginJsonOutput)
            res = json.loads(res)
        # Ignore nondeterministic outputs
        res.pop("date_start")
        res.pop("date_end")
        res.pop("runtime")
        print(res["events"])
        self.assertEqual(
            res,
            {
                "state": {"label": "completed"},
                "feature_types": [{"desc": "A return value", "name": "value", "type": "integer"}],
                "events": [
                    {
                        "sha256": sha256,
                        "data": [
                            {
                                "hash": "1c2568562238a338919f87598737756a67e22f35d5031a45796c9d312b5beb54",
                                "label": "text",
                                "language": "python",
                            }
                        ],
                        "features": {"value": [{"value": "12345"}]},
                    }
                ],
                "data": {
                    "1c2568562238a338919f87598737756a67e22f35d5031a45796c9d312b5beb54": "VGV4dCBvdXRwdXQgY29udGVudA==",
                },
            },
        )

    class TestPluginOutputFolder(MainBaseTestPlugin):
        VERSION = "1.0"
        FEATURES = [Feature("value", "A return value", type=FeatureType.Integer)]

        def execute(self, job: Job) -> dict:
            self.add_data(
                "text",
                {"language": "python"},
                b"Text output content",
            )
            self.add_feature_values("value", 12345)

    def test_output_folder(self):
        """Test that the plugin can output artifacts and run info to folder."""
        with tempfile.NamedTemporaryFile() as f, tempfile.TemporaryDirectory() as tmpdir:
            f.write(b"blah")
            f.seek(0)
            sys.argv = ["test_main.py", f.name, "--output-folder", tmpdir]
            self._stdout_result(self.TestPluginOutputFolder)
            self.assertEqual(
                os.listdir(tmpdir),
                ["1c2568562238a338919f87598737756a67e22f35d5031a45796c9d312b5beb54_text.data"],
            )
            with open(
                os.path.join(tmpdir, "1c2568562238a338919f87598737756a67e22f35d5031a45796c9d312b5beb54_text.data"),
                "rb",
            ) as f2:
                databytes = f2.read()
            self.assertEqual(databytes, b"Text output content")

    class TestPluginMultipluginRegular(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [Feature("value", "A return value", FeatureType.String)]

        def __init__(self, config: dict[str, dict[str, Any]] = None) -> None:
            super().__init__(config)
            self.register_multiplugin("66_dec", "version 5", self.execute_order_66)
            self.register_multiplugin("rol_67", "version 5", self.execute_order_67)

        def execute_order_66(self, entity) -> dict:
            return {"features": {"value": "for a safe and secure galaxy"}}

        def execute_order_67(self, entity) -> dict:
            return {"features": {"value": "hello there"}}

    def test_multiplugin_regular(self):
        # First, execute without input data streams
        sys.argv = ["test_main.py"]
        res = self._stdout_result(self.TestPluginMultipluginRegular)
        expected = """
----- TestPlugin-66_dec results -----
COMPLETED

Output features:
  value: for a safe and secure galaxy


Feature key:
  value:  A return value

----- TestPlugin-rol_67 results -----
COMPLETED

Output features:
  value: hello there


Feature key:
  value:  A return value
        """.strip()
        self.assertEqual(res.strip(), expected)

    class TestPluginMultipluginRegular(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [Feature("value", "A return value", FeatureType.String)]

        def __init__(self, config: dict[str, dict[str, Any]] = None) -> None:
            super().__init__(config)
            self.register_multiplugin("66_dec", "version 5", self.execute_order_66)
            self.register_multiplugin("rol_67", "version 5", self.execute_order_67)

        def execute(self, job: Job) -> dict:
            pass

        def execute_order_66(self, job: Job) -> dict:
            self.add_feature_values("value", "for a safe and secure galaxy")

        def execute_order_67(self, job: Job) -> dict:
            self.add_feature_values("value", "hello there")

    def test_multiplugin_regular(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"blah")
            f.seek(0)
            sha256 = hashlib.sha256(f.read()).hexdigest()
            f.seek(0)
            sys.argv = ["test_main.py", f.name, "--output-json"]
            results = self._stdout_result(self.TestPluginMultipluginRegular)
        res1, res2, _, _ = results.split("\n")

        res1 = json.loads(res1)
        # Ignore nondeterministic outputs
        res1.pop("date_start")
        res1.pop("date_end")
        res1.pop("runtime")
        print(res1)
        self.assertEqual(
            res1,
            {
                "state": {"label": "completed"},
                "feature_types": [{"name": "value", "desc": "A return value", "type": "string"}],
                "events": [
                    {
                        "sha256": sha256,
                        "features": {
                            "value": [
                                {
                                    "value": "for a safe and secure galaxy",
                                }
                            ]
                        },
                    }
                ],
            },
        )

        res2 = json.loads(res2)
        # Ignore nondeterministic outputs
        res2.pop("date_start")
        res2.pop("date_end")
        res2.pop("runtime")
        print(res2)
        self.assertEqual(
            res2,
            {
                "state": {"label": "completed"},
                "feature_types": [{"name": "value", "desc": "A return value", "type": "string"}],
                "events": [
                    {
                        "sha256": sha256,
                        "features": {"value": [{"value": "hello there"}]},
                    }
                ],
            },
        )

    class TestPluginFolder(MainBaseTestPlugin):
        VERSION = "none"

        def execute(self, job: Job) -> dict:
            self.add_data(
                "text",
                {"language": "python"},
                b"Text output content",
            )
            return

    def test_folder(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            pf1 = os.path.join(tmpdir, "f1")
            pf2 = os.path.join(tmpdir, "f2")
            with open(pf1, "w+b") as f:
                f.write(b"data1")
                f.seek(0)
                sha2561 = hashlib.sha256(f.read()).hexdigest()

            with open(pf2, "w+b") as f:
                f.write(b"data2")
                f.seek(0)
                sha2562 = hashlib.sha256(f.read()).hexdigest()
            sys.argv = ["test_main.py", tmpdir]
            res = self._stdout_result(self.TestPluginFolder)
        print(res)
        self.assertEqual(
            res,
            f"""{pf1}
----- TestPluginFolder results -----
COMPLETED

events (1)

event for {sha2561}:None
  {{}}
  output data streams (1):
    19 bytes - EventData(hash='1c2568562238a338919f87598737756a67e22f35d5031a45796c9d312b5beb54', label='text', language='python')

{pf2}
----- TestPluginFolder results -----
COMPLETED

events (1)

event for {sha2562}:None
  {{}}
  output data streams (1):
    19 bytes - EventData(hash='1c2568562238a338919f87598737756a67e22f35d5031a45796c9d312b5beb54', label='text', language='python')

""",
        )

    class TestPluginFolderJson(MainBaseTestPlugin):
        VERSION = "1.0"
        FEATURES = [Feature("value", "A return value", type=FeatureType.Integer)]

        def execute(self, job: Job) -> dict:
            self.add_data(
                "text",
                {"language": "python"},
                b"Text output content",
            )
            self.add_feature_values("value", 12345)

    def test_folder_json(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            pf1 = os.path.join(tmpdir, "f1")
            pf2 = os.path.join(tmpdir, "f2")
            with open(pf1, "w+b") as f:
                f.write(b"data1")
                f.seek(0)
                sha2561 = hashlib.sha256(f.read()).hexdigest()

            with open(pf2, "w+b") as f:
                f.write(b"data2")
                f.seek(0)
                sha2562 = hashlib.sha256(f.read()).hexdigest()
            sys.argv = ["test_main.py", tmpdir, "--output-json"]
            res = self._stdout_result(self.TestPluginFolderJson)

        # every second line is valid json
        lines = res.split("\n")
        line1 = json.loads(lines[1])
        line2 = json.loads(lines[3])
        # Ignore nondeterministic outputs
        line1.pop("date_start")
        line1.pop("date_end")
        line1.pop("runtime")
        print(line1)
        self.assertEqual(
            line1,
            {
                "state": {"label": "completed"},
                "feature_types": [{"name": "value", "desc": "A return value", "type": "integer"}],
                "events": [
                    {
                        "sha256": sha2561,
                        "data": [
                            {
                                "hash": "1c2568562238a338919f87598737756a67e22f35d5031a45796c9d312b5beb54",
                                "label": "text",
                                "language": "python",
                            }
                        ],
                        "features": {"value": [{"value": "12345"}]},
                    }
                ],
                "data": {
                    "1c2568562238a338919f87598737756a67e22f35d5031a45796c9d312b5beb54": "VGV4dCBvdXRwdXQgY29udGVudA==",
                },
            },
        )
        # Ignore nondeterministic outputs
        line2.pop("date_start")
        line2.pop("date_end")
        line2.pop("runtime")
        print(line2)
        self.assertEqual(
            line2,
            {
                "state": {"label": "completed"},
                "feature_types": [{"name": "value", "desc": "A return value", "type": "integer"}],
                "events": [
                    {
                        "sha256": sha2562,
                        "data": [
                            {
                                "hash": "1c2568562238a338919f87598737756a67e22f35d5031a45796c9d312b5beb54",
                                "label": "text",
                                "language": "python",
                            }
                        ],
                        "features": {"value": [{"value": "12345"}]},
                    }
                ],
                "data": {
                    "1c2568562238a338919f87598737756a67e22f35d5031a45796c9d312b5beb54": "VGV4dCBvdXRwdXQgY29udGVudA==",
                },
            },
        )

    class TestPluginServerExceptionHandlingPlugin(MainBaseTestPlugin):
        VERSION = "none"
        FEATURES = [Feature("value", "A return value", type=FeatureType.Integer)]

        def execute(self, job: Job) -> dict:
            self.add_feature_values("value", 97)

    def test_server_exception_handling_plugin(self):
        """Tests the server's handling of errors from coordinator."""
        # Have to use shared memory because it's the only way for the mock method to communicate from the subprocess.
        # back to the test, to validate the right number of calls were made.
        with managers.SharedMemoryManager() as smm:
            share_mem = smm.ShareableList([0])

            # Normal exit
            def fake_run_loop_return(*args, **kwargs):
                share_mem[0] += 1
                return

            with mock.patch.object(monitor.Monitor, "run_loop", wraps=fake_run_loop_return):
                main.execute(self.TestPluginServerExceptionHandlingPlugin, main.Args())
                self.assertEqual(share_mem[0], 1)

            share_mem[0] = 0

            # Exception exit
            def fake_run_loop_exception(*args, **kwargs):
                share_mem[0] += 1
                raise Exception()

            with mock.patch.object(monitor.Monitor, "run_loop", wraps=fake_run_loop_exception):
                with self.assertRaises(Exception):
                    main.execute(self.TestPluginServerExceptionHandlingPlugin, main.Args())
                    self.assertEqual(share_mem[0], 1)

            # Re-boot plugin once and cleanup temp
            share_mem[0] = 0

            def fake_run_loop_retry(*args, **kwargs):
                share_mem[0] += 1
                if share_mem[0] == 1:
                    raise monitor.RecreateException()
                else:
                    return

            with mock.patch.object(monitor.Monitor, "run_loop", wraps=fake_run_loop_retry):
                with self.assertRaises(monitor.RecreateException):
                    main.execute(self.TestPluginServerExceptionHandlingPlugin, main.Args())
                    self.assertEqual(share_mem[0], 2)
